import statistics


class Segments(list):
    """
    A list of [level, time] pairs.
    """
    def __init__(self, segments=(), length=None):
        super().__init__()
        self.base_value = 0
        for segment in segments:
            try:
                level, time = segment
            except TypeError:
                level, time = segment, None
            self.append([level, time])

        times = [t for s, t in self if t is not None]
        if times:
            mean = sum(times) / len(times)
        else:
            mean = (length or 1) / max(1, len(self))

        for segment in self:
            if segment[1] is None:
                segment[1] = mean

        self.total_time = sum(t for l, t in self)

    def __call__(self, time, base_value=0):
        elapsed_time = 0
        level = base_value
        for l, t in self:
            segment_end_time = elapsed_time + t
            if time < segment_end_time:
                delta_t = time - elapsed_time
                return level + (l - level) * delta_t / t

            elapsed_time = segment_end_time
            level = l

        return level



import pathlib
from . import rest_server
from .. import control
from ... util import server_cache

PORT = 8787
ROOT_FOLDER = pathlib.Path(__file__).parents[3] / 'ui' / 'rest'
INDEX_FILE = 'index.html'


class Rest(control.ExtractedControl, server_cache.StaticCache):
    OPEN_PAGE = False
    SERVER_CLASS = rest_server.RestServer

    def __init__(self, *args, port=PORT, external_access=False, open_page=None,
                 root_folder=ROOT_FOLDER, index_file=INDEX_FILE, **kwds):
        super().__init__(*args, **kwds)
        if open_page is None:
            open_page = self.OPEN_PAGE

        cached_server = self.cache().get_server(
            port, external_access=external_access,
            open_page=open_page, root_folder=root_folder, index_file=index_file)
        self.rest_server = cached_server.server

    def set_project(self, project):
        self.rest_server.set_project(project)
        super().set_project(project)

    def cleanup(self):
        self.rest_server.set_project(None)
        super().cleanup()

    def _make_thread(self):
        return self.rest_server


class OpenPage(Rest):
    OPEN_PAGE = True



"""
Create a new BLiPS project
"""

import os, re, string
from .. util import log

DESCRIPTION = """
Example:

.. code-block: bash

    # Create a new project directory named my_project/ in the current directory
    bp new my_project

    # Create a new project directory named my_project/ in ~/projects
    bp new my_project ~/projects

"""

PUNCTUATION = '._'
LEGAL_CHARS = set(string.ascii_letters + string.digits + PUNCTUATION)
DIR_NAME = os.path.dirname(__file__)


def run(args):
    name = args.project_name
    bad = set(name) - LEGAL_CHARS
    if bad:
        raise ValueError('Bad characters in project: "%s"' % ''.join(bad))
    if not name[0].isalpha():
        raise ValueError('Project names must start with a character')

    directory = args.directory or name
    if os.path.exists(directory):
        raise ValueError('Directory "%s" already exists' % directory)

    os.makedirs(directory)
    words = _split_words(name)
    class_name = ''.join(w.capitalize() for w in words)
    py_name = name.replace('.', '_').replace('-', '_')

    context = {
        'class_name': class_name,
        'name': name,
        'py_name': py_name,
    }

    for template, out_file in TEMPLATES.items():
        out_file = out_file.format(**context)
        out_file = os.path.join(directory, out_file)
        tmpl_file = os.path.join(DIR_NAME, template)
        tmpl_data = open(tmpl_file).read()

        with open(out_file, 'w') as fp:
            fp.write(tmpl_data.format(**context))
            log.printer('Written', out_file)

    log.printer('Created new project in', directory)


def add_arguments(parser):
    parser.set_defaults(run=run)
    parser.add_argument('project_name', help=PROJECT_NAME_HELP)
    parser.add_argument('directory', default='', nargs='?', help=DIRECTORY_HELP)


TEMPLATES = {
    'template/project.yml.tmpl': '{name}.yml',
    'template/animation.py.tmpl': '{py_name}.py',
}

PROJECT_NAME_HELP = """
The Name of the new BLiPS project you want to create.

A Project Name can contain letters, numbers, or `_`. Nothing else is
allowed, which means no whitespace, "no `/`" and no `.`.

A project name must start with a letter.

Valid project names are:

* P
* project_name
* IHeartArea51

Invalid project names are:

* 23skidoo
* project.name
* _23skidoo

"""

DIRECTORY_HELP = """
Optional directory where you want to create the project.
If absent, use the name of the project as the directory name.
"""

_split_words = re.compile(r'''
    # Find words in a string. Order matters!
    [A-Z]+(?=[A-Z][a-z]) |  # All upper case before a capitalized word
    [A-Z]?[a-z]+ |  # Capitalized words / all lower case
    [A-Z]+ |  # All upper case
    \d+  # Numbers
''', re.VERBOSE).findall
# https://stackoverflow.com/a/41510011/43839



import platform, subprocess

MAC = 'Darwin'
WINDOWS = 'Windows'
CPUINFO_FILE = '/proc/cpuinfo'


class Platform:
    def __init__(self):
        self.platform = platform.system()
        self.version = platform.version()
        self.release = platform.release()
        self.python_version = platform.python_version()

        try:
            self.cpuinfo = [i.strip() for i in open(CPUINFO_FILE)]
        except:
            self.cpuinfo = []

        def is_rpi_line(i):
            return i.startswith('Hardware') and i.endswith('BCM2708')

        self.is_raspberry_pi = any(is_rpi_line(i) for i in self.cpuinfo)
        self.is_linux = (self.platform == 'linux')

        platform_version = ()
        if self.is_linux:
            # Use the linux distribution as the name
            self.platform = platform.linux_distribution()[0].lower()
        elif self.platform == WINDOWS:
            platform_version = platform.win32_ver()
        elif self.platform == MAC:
            release, versioninfo, machine = platform.mac_ver()
            platform_version = release, machine
            # https://boklee.blogspot.com/2012/05/how-to-retrieve-cpuinfo-on-os-x.html
            for i in 'features', 'brand_string':
                s = subprocess.check_output(('sysctl', 'machdep.cpu.' + i))
                self.cpuinfo.append(s.decode().strip())

        self.platform_version = ':'.join(platform_version)



from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from . import interfaces
from enum import IntEnum
from .. ledtype import LEDTYPE
# SubDriver imports
from . APA102 import APA102
from . LPD8806 import LPD8806
from . WS281X import WS281X
from . WS2801 import WS2801


# This may look weird, but was done to keep the IDs the same
# as Serial so that ledtype can always be resolved the same
SPI_DRIVERS = {
    1: LPD8806,
    2: WS2801,
    3: WS281X,
    9: APA102
}


def SPI(ledtype=None, num=0, **kwargs):
    """Wrapper function for using SPI device drivers on systems like the
    Raspberry Pi and BeagleBone. This allows using any of the SPI drivers
    from a single entry point instead importing the driver for a specific
    LED type.

    Provides the same parameters of
    :py:class:`bibliopixel.drivers.SPI.SPIBase` as
    well as those below:

    :param ledtype: One of: LPD8806, WS2801, WS281X, or APA102
    """

    from ...project.types.ledtype import make
    if ledtype is None:
        raise ValueError('Must provide ledtype value!')
    ledtype = make(ledtype)

    if num == 0:
        raise ValueError('Must provide num value >0!')
    if ledtype not in SPI_DRIVERS.keys():
        raise ValueError('{} is not a valid LED type.'.format(ledtype))

    return SPI_DRIVERS[ledtype](num, **kwargs)



import os, sys
from . import merge as _merge
from .. util import data_file, log, persistent_dict

# This is set to true during testing.
BYPASS_PROJECT_DEFAULTS = False

USER_DEFAULTS_FILE = os.path.expanduser('~/.bibliopixel/default.json')
USER_DEFAULTS = persistent_dict.PersistentDict(USER_DEFAULTS_FILE)
PROJECT_DEFAULTS = []

SAVE_DIRECTORY = os.path.expanduser('~/.bibliopixel/project_defaults')


def merge(*projects):
    """
    Merge the global project defaults and the user project defaults with a list
    of projects
    """
    defaults = [_merge.DEFAULT_PROJECT]
    if not BYPASS_PROJECT_DEFAULTS:
        defaults += PROJECT_DEFAULTS or [USER_DEFAULTS]

    return _merge.merge(*(defaults + list(projects)))


def show_defaults():
    """List current user defaults in data_file format (JSON or YAML)"""
    _warn_if_empty()
    data_file.dump(USER_DEFAULTS, sys.stdout)
    log.printer()


def reset_defaults(sections):
    """Reset some or all sections of the user defaults file"""
    if _warn_if_empty():
        return
    sections = sorted(_check_sections(sections) or USER_DEFAULTS)

    log.printer('Before reset:')
    show_defaults()

    for s in sections:
        try:
            del USER_DEFAULTS[s]
        except:
            pass

    log.printer('Reset', *sections)


def set_defaults(sections):
    """
    Set user default values for one or more sections, or for an entire project
    at once.
    """
    assignments = _sections_to_assignments(sections)
    USER_DEFAULTS.update(assignments)


def load_defaults(name):
    defaults = _load_defaults(name)
    USER_DEFAULTS.clear()
    USER_DEFAULTS.update(defaults)
    log.printer('Loaded project defaults from', name)


def save_defaults(name):
    if '/' in name or '.' in name:
        path = name
    else:
        os.makedirs(SAVE_DIRECTORY, exist_ok=True)
        path = _default_file(name)

    if os.path.exists(path):
        yn = input('Default %s already exists.  Overwrite? (y/N) ' % name)
        if not yn.lower().startswith('y'):
            return

    with open(path, 'w') as fp:
        data_file.dump(USER_DEFAULTS, fp)

    log.printer('Written project defaults to', name)


def remove_defaults(name):
    if name in _defaults():
        os.remove(_default_file(name))
    else:
        log.printer('No such default:', name)
        list_saved_defaults()


def list_saved_defaults():
    defaults = _defaults()
    if defaults:
        log.printer('Saved project defaults:')
        for i in defaults:
            log.printer('   ', i)
    else:
        log.printer('(no project defaults saved)')


def set_project_defaults(names):
    PROJECT_DEFAULTS[:] = [_load_defaults(n) for n in names]


def _check_sections(sections):
    unknown = set(sections) - set(_merge.PROJECT_SECTIONS)
    if not unknown:
        return sections

    s = '' if len(unknown) == 1 else 's'
    unknown = ' '.join(sorted(unknown))
    raise ValueError('Unknown project section%s: %s' % (s, unknown))


# Fail immediately if the user preferences contain unknown sections.
# This should only ever happen if the user edited the file by hand.
_check_sections(USER_DEFAULTS)


def _sections_to_assignments(sections):
    if not sections:
        raise ValueError('`set` called with no arguments')

    is_json = sections[0].startswith('{')
    if is_json:
        return _check_sections(data_file.loads(' '.join(sections)))

    assignments = {}
    for s in sections:
        try:
            section, value = s.split('=', 1)
        except:
            raise ValueError('All arguments to `set` must contain =')

        # Help the user out a little if they use strings without quotes by
        # wrapping them in quotes.

        if value.isalpha() and (value not in ('true', 'false', 'null')):
            value = '"%s"' % value

        try:
            value = data_file.loads(value)
        except Exception as e:
            e.args += ('Bad data in `set` for %s=%s' % (section, value),)
            raise

        *first, last = section.split('.')
        a = assignments
        for i in first:
            a = a.setdefault(i, [] if i.endswith('s') else {})
        a[last] = value

    return _check_sections(assignments)


def _warn_if_empty():
    if not USER_DEFAULTS:
        log.printer('(no entries in defaults file)', file=sys.stderr)
        return True


def _defaults():
    if not os.path.exists(SAVE_DIRECTORY):
        return []
    return [f for f in os.listdir(SAVE_DIRECTORY) if not f.startswith('.')]


def _default_file(name):
    return os.path.join(SAVE_DIRECTORY, name)


def _load_defaults(name):
    n = name
    if not os.path.exists(n):
        n = _default_file(n)
        if not os.path.exists(n):
            list_saved_defaults()
            raise ValueError('No such default: ' + name)
    try:
        return data_file.load(n)
    except Exception as e:
        e.args = ('There was an error in data file ' + name,
                  'Did you edit this file by hand?') + e.args
        raise



class ChannelOrder(object):
    """
    Helper class to automatically convert string values into tuple
    values needed to define color channel order.
    """

    RGB = 0, 1, 2
    RBG = 0, 2, 1
    GRB = 1, 0, 2
    GBR = 1, 2, 0
    BRG = 2, 0, 1
    BGR = 2, 1, 0

    ORDERS = RGB, RBG, GRB, GBR, BRG, BGR

    @staticmethod
    def make(x):
        orders = ChannelOrder.ORDERS
        try:
            length = len(x)
        except:
            return orders[x % len(orders)]

        if length != 3:
            raise ValueError('ChannelOrder "%s" does not have length 3' % x)

        try:
            s = x.lower()
        except AttributeError:
            s = tuple(x)
        else:
            try:
                s = tuple('rgb'.index(i) for i in s)
            except ValueError:
                raise ValueError('ChannelOrder "%s" has non-rgb elements' % x)

        if not all(0 <= x <= 2 for x in s):
            raise ValueError(
                'ChannelOrder %s has members not between 0 and 2.' % x)

        try:
            # Canonicalize and detect dupes at the same time.
            return orders[orders.index(s)]
        except:
            raise ValueError('ChannelOrder %s has duplicate elements' % x)



from bibliopixel.drivers.SimPixel import SimPixel
from bibliopixel import Circle, Matrix
from bibliopixel.layout import geometry
from bibliopixel.animation import Sequence
from BiblioPixelAnimations.circle.bloom import CircleBloom
from BiblioPixelAnimations.circle.swirl import Swirl
from BiblioPixelAnimations.circle.hyperspace import HyperspaceRainbow


BLOOM = {
    'driver': {
        'typename': 'simpixel',
        'num': 0
    },

    'layout': {
        'typename': 'matrix',
        'width': 0,
        'height': 0,
    },

    'animation': {
        'typename': 'BiblioPixelAnimations.matrix.bloom.Bloom'
    },

    'run': {
        'amt': 6,
        'fps': 30
    }
}


MATRIX_PROJECT = {
    'driver': {
        'typename': 'simpixel',
        'num': 0
    },

    'layout': {
        'typename': 'matrix',
        'width': 0,
        'height': 0,
    },

    'animation': {
        'typename': 'sequence',
        'animations': [
            {
                'animation':
                'BiblioPixelAnimations.matrix.MatrixRain.MatrixRainBow',
                'run': {
                    'amt': 1,
                    'fps': 20,
                    'seconds': 8,
                },
            },
            {
                'animation': {
                    'typename': 'BiblioPixelAnimations.matrix.Text.ScrollText',
                    'xPos': 16,
                    'font_scale': 2,
                    'text': 'BiblioPixel Demo'
                },
                'run': {
                    'amt': 1,
                    'fps': 20,
                    'seconds': 8,
                    'until_complete': True,
                },
            },
            {
                'animation': 'BiblioPixelAnimations.matrix.bloom.Bloom',
                'run': {
                    'amt': 3,
                    'fps': 60,
                    'seconds': 8,
                },
            },
            {
                'animation': 'BiblioPixelAnimations.matrix.circlepop.CirclePop',
                'run': {
                    'amt': 1,
                    'fps': 15,
                    'seconds': 8,
                },
            },
        ],
    },
}

CUBE_PROJECT = {
    'driver': {
        'typename': 'simpixel',
        'num': 0
    },

    'layout': {
        'typename': 'cube',
        'x': 0,
        'y': 0,
        'z': 0
    },

    'animation': {
        'typename': 'sequence',
        'animations': [
            {
                'animation': {
                    'typename':
                        'BiblioPixelAnimations.cube.wave_spiral.WaveSpiral',
                    'dir': False,
                    'offset': 6
                },
                'run': {
                    'amt': 1,
                    'fps': 15,
                    'seconds': 8
                },
            },
            {
                'animation': 'BiblioPixelAnimations.cube.Rain.RainBow',
                'run': {
                    'amt': 1,
                    'fps': 10,
                    'seconds': 8
                },
            },
            {
                'animation': 'BiblioPixelAnimations.cube.bloom.CubeBloom',
                'run': {
                    'amt': 6,
                    'fps': 20,
                    'seconds': 8,
                },
            }
        ],
    },
}

_PIXELS_PER = [1, 4, 8, 12, 18, 24, 32, 40, 52, 64]
_RINGS, _STEPS = geometry.make_circle_coord_map(pixels_per=_PIXELS_PER)
_POINTS = geometry.make_circle_coord_map_positions(
    _RINGS, origin=(200, 200, 0), z_diff=16)

CIRCLE_PROJECT = {
    'driver': {
        'typename': 'simpixel',
        'num': sum(_PIXELS_PER),
        'pixel_positions': _POINTS,
    },

    'layout': {
        'typename': 'circle',
        'rings': _RINGS,
        'maxAngleDiff': 0,
    },

    'animation': {
        'typename': 'sequence',
        'length': 8,
        'animations': [CircleBloom, Swirl, HyperspaceRainbow]
    }
}


def circle(args):
    pixels_per = [1, 4, 8, 12, 18, 24, 32, 40, 52, 64]
    rings, steps = geometry.make_circle_coord_map(pixels_per=pixels_per)
    points = geometry.make_circle_coord_map_positions(
        rings, origin=(200, 200, 0), z_diff=16)
    driver = SimPixel(sum(pixels_per), pixel_positions=points)
    layout = Circle(driver, rings=rings, maxAngleDiff=0)
    anim = Sequence(layout)

    anim.add_animation(CircleBloom(layout), amt=3, fps=30, seconds=8)
    anim.add_animation(Swirl(layout, angle=4), amt=6, fps=15, seconds=8)
    anim.add_animation(HyperspaceRainbow(layout), fps=15, seconds=8)

    return anim


DEMO_TABLE = {
    'bloom': BLOOM,
    'circle': CIRCLE_PROJECT,
    'cube': CUBE_PROJECT,
    'matrix': MATRIX_PROJECT,
}



"""
Monitor a control source
"""

import sys
from .. project import importer
from .. util import log


def run(args):
    control = args.control[0]
    try:
        tc = importer.import_symbol(control, 'bibliopixel.control')
    except:
        log.error('Do not understand control "%s"', control)
        raise

    control_object = tc(pre_routing='()')
    control_object.set_root(log.printer)

    control_object.start()
    control_object.wait()


def add_arguments(parser):
    parser.add_argument(
        'control', nargs=1,
        help='Name of a control to monitor')

    parser.set_defaults(run=run)



"""
Print information about BiblioPixel
"""

DESCRIPTION = """
Prints the versions of BiblioPixel's dependencies, and the platform
that the program is running on.
"""

import datetime, os, platform, sys
import bibliopixel, BiblioPixelAnimations, loady
from bibliopixel.util import log
from bibliopixel.util.platform import Platform

NONE = '(none)'
MODULES = bibliopixel, BiblioPixelAnimations, loady


def run(args):
    platform = Platform()

    now = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')
    bp_path = sys.argv[0]
    library_path = os.path.dirname(bibliopixel.__file__)
    dependencies = '\n'.join(_dependency(m) for m in MODULES)
    if platform.cpuinfo:
        cpuinfo = CPUINFO % '\n'.join(platform.cpuinfo)
    else:
        cpuinfo = ''

    log.printer(MESSAGE.format(**locals()))


def add_arguments(parser):
    parser.set_defaults(run=run)


def _dependency(module):
    path = os.path.dirname(module.__file__)
    parent = os.path.dirname(path)

    try:
        fp = open(os.path.join(path, 'VERSION'))
    except:
        try:
            fp = open(os.path.join(parent, 'VERSION'))
        except:
            fp = None

    version = fp.read().strip() if fp else NONE

    try:
        import git
        repo = git.Repo(os.path.dirname(path))
    except:
        commit_id = tag = NONE
    else:
        commit_id = repo.commit('HEAD').hexsha[:7]
        tag = repo.tags[-1].name if repo.tags else '(none)'

    return ('    %s: version %s, git commit: %s, git tag %s' %
            (module.__name__, version, commit_id, tag))


MESSAGE = """\
Timestamp:        {now}
Python version:   {platform.python_version}
`bp` path:        {bp_path}
Library path:     {library_path}

Platform:         {platform.platform}
Platform version: {platform.platform_version}
Platform release: {platform.release}

Dependencies:
{dependencies}
{cpuinfo}"""

CPUINFO = """
cpuinfo
--------------------------------------------------------------------------------
%s
--------------------------------------------------------------------------------
"""



from .. util import AttributeDict


class BaseGamePad(object):
    def __init__(self):
        pass

    def getKeys(self):
        raise Exception("getKeys must be overriden!")

    def close(self):
        pass

    def setLights(self, data):
        pass

    def setLightsOff(self, count):
        pass



from . import construct, load

"""
In order to validate project descriptions, we need to recurse through project
dictionaries before we actually construct the objects in them.

Some fields of project dictionaries are designated to be "type constructors" -
i.e. they describe how to create an object.  We need to recurse down through
those - and how we do so depends on the type of the object.

For example, at the top level, the "layout" and "animation" fields are
type constructors while the "run" and "path" fields are not.

So recurse uses static class members on the object class being created
to determine what to do before and after recursion, and how to compute
children.
"""


def recurse(desc, pre='pre_recursion', post=None, python_path=None):
    """
    Depth first recursion through a dictionary containing type constructors

    The arguments pre, post and children are independently either:

    * None, which means to do nothing
    * a string, which means to use the static class method of that name on the
      class being constructed, or
    * a callable, to be called at each recursion

    Arguments:

    dictionary -- a project dictionary or one of its subdictionaries
    pre -- called before children are visited node in the recursion
    post -- called after children are visited in the recursion
    python_path -- relative path to start resolving typenames

    """
    def call(f, desc):
        if isinstance(f, str):
            # f is the name of a static class method on the datatype.
            f = getattr(datatype, f, None)
        return f and f(desc)

    # Automatically load strings that look like JSON or Yaml filenames.
    desc = load.load_if_filename(desc) or desc

    desc = construct.to_type_constructor(desc, python_path)
    datatype = desc.get('datatype')

    desc = call(pre, desc) or desc

    for child_name in getattr(datatype, 'CHILDREN', []):
        child = desc.get(child_name)
        if child:
            is_plural = child_name.endswith('s')
            remove_s = is_plural and child_name != 'drivers'
            # This is because it's the "drivers" directory, whereas
            # the others are animation, control, layout, project
            # without the s. TODO: rename drivers/ to driver/ in v4

            cname = child_name[:-1] if remove_s else child_name
            new_path = python_path or ('bibliopixel.' + cname)
            if is_plural:
                if isinstance(child, (dict, str)):
                    child = [child]
                for i, c in enumerate(child):
                    child[i] = recurse(c, pre, post, new_path)
                desc[child_name] = child
            else:
                desc[child_name] = recurse(child, pre, post, new_path)

    d = call(post, desc)
    return desc if d is None else d



import flask, functools, traceback, urllib
from .. import editor

NO_PROJECT_ERROR = 'No Project is currently loaded'
BAD_ADDRESS_ERROR = 'Bad address {address}'
BAD_GETTER_ERROR = 'Couldn\'t get address {address}'
BAD_SETTER_ERROR = 'Couldn\'t set value {value} at address {address}'


def single(method):
    """Decorator for RestServer methods that take a single address"""
    @functools.wraps(method)
    def single(self, address, value=None):
        address = urllib.parse.unquote_plus(address)
        try:
            error = NO_PROJECT_ERROR
            if not self.project:
                raise ValueError
            error = BAD_ADDRESS_ERROR
            ed = editor.Editor(address, self.project)

            if value is None:
                error = BAD_GETTER_ERROR
                result = method(self, ed)
            else:
                error = BAD_SETTER_ERROR
                result = method(self, ed, value)
            result = {'value': result}

        except Exception as e:
            traceback.print_exc()
            msg = '%s\n%s' % (error.format(**locals()), e)
            result = {'error': msg}

        return flask.jsonify(result)

    return single


def multi(method):
    """Decorator for RestServer methods that take multiple addresses"""
    @functools.wraps(method)
    def multi(self, address=''):
        values = flask.request.values
        address = urllib.parse.unquote_plus(address)
        if address and values and not address.endswith('.'):
            address += '.'

        result = {}
        for a in values or '':
            try:
                if not self.project:
                    raise ValueError('No Project is currently loaded')

                ed = editor.Editor(address + a, self.project)
                result[address + a] = {'value': method(self, ed, a)}
            except:
                if self.project:
                    traceback.print_exc()
                result[address + a] = {'error': 'Could not multi addr %s' % a}

        return flask.jsonify(result)

    return multi



import itertools
from .. import data_file

MODULE = '.'.join(__name__.split('.')[:-1])

YML_READER_TYPE = MODULE + '.yml.Reader'
SEPARATOR = '---\n'
COLORS_PER_LINE = 12


class Writer:
    def __init__(self, movie_writer):
        self.shape = movie_writer.project.layout.shape
        self.ranges = tuple(range(i) for i in self.shape)
        self.getter = movie_writer.project.layout.get
        self.movie_writer = movie_writer
        self.fp = None

    def step(self):
        if not self.fp:
            self.fp = open(self.movie_writer.filename, 'w')
            header = {
                'typename': YML_READER_TYPE,
                'fps': self.movie_writer.scaled_fps,
                'shape': self.shape,
            }
            data_file.dump(header, self.fp)

        self.fp.write(SEPARATOR)
        self.fp.write('"')
        for i, index in enumerate(itertools.product(*self.ranges)):
            if not (i % COLORS_PER_LINE):
                self.fp.write('\\\n')
            self.fp.write('%02x%02x%02x' % self.getter(*index))
        self.fp.write('"\n')

    def write(self):
        try:
            self.fp.close()
        except:
            pass



import itertools
from . import names


def closest_colors(color_name, metric=None):
    metric = metric or euclidean
    color = names.to_color(color_name)

    mn_list = [(metric(color, c), n) for n, c in names.COLOR_DICT.items()]
    mn_list.sort()
    min_metric = mn_list[0][0]
    result = []
    for met, name in mn_list:
        if met > min_metric:
            return sorted(result)
        result.append(name)


def euclidean(c1, c2):
    """Square of the euclidean distance"""
    diffs = ((i - j) for i, j in zip(c1, c2))
    return sum(x * x for x in diffs)


def taxicab(c1, c2):
    diffs = ((i - j) for i, j in zip(c1, c2))
    return sum(abs(x) for x in diffs)


def all_colors(start=0, stop=255, skip=1):
    return itertools.product(range(start, stop, skip), repeat=3)



import os, queue, threading, time, weakref
from . import (
    attributes, construct, fill, defaults, edit_queue, load, recurse)
from .. util import exception, log

EDIT_QUEUE_MAXSIZE = 1000
NO_DATATYPE_ERROR = 'No "datatype" field in section "%s"'
MIN_TIME = 0.0001


class Project:
    CHILDREN = 'maker', 'drivers', 'layout', 'animation', 'controls'
    LOCK = threading.Lock()

    @staticmethod
    def pre_recursion(desc):
        return fill.fill(desc)

    def construct_child(self, section_name, datatype=None, typename=None,
                        **kwds):
        if not datatype:
            raise ValueError(NO_DATATYPE_ERROR % section_name)

        construct = getattr(datatype, 'construct', None)
        if construct:
            return construct(self, **kwds)
        return datatype(**kwds)

    def __init__(self, *,
                 drivers, layout, maker, path, animation, controls,
                 edit_queue_maxsize=EDIT_QUEUE_MAXSIZE, **kwds):
        """
        :param int edit_queue_maxsize: maxsize parameter to queue.Queue.
            0 means an unbounded queue.
        """
        def create(root, name):
            def post(desc):
                exception = desc.get('_exception')
                if exception:
                    raise exception
                return self.construct_child(name, **desc)

            with exception.add('Unable to create ' + name):
                return recurse.recurse(
                    root,
                    pre=None,
                    post=post,
                    python_path='bibliopixel.' + name)

        attributes.check(kwds, 'project')
        self.path = path
        layout = layout or fill.fill_layout(animation)
        self.maker = self.construct_child('maker', **maker)
        self.drivers = [create(d, 'drivers') for d in drivers]
        with exception.add('Unable to create layout'):
            self.layout = self.construct_child('layout', **layout)

        self.animation = create(animation, 'animation')
        self.animation.top_level = True
        self.running = False
        self.clock = time

        eq = edit_queue.EditQueue(maxsize=edit_queue_maxsize)
        self.layout.edit_queue = self.animation.edit_queue = eq
        self.animation.add_preframe_callback(eq.get_and_run_edits)

        # Unfortunately, the whole animation cycle is controlled by methods on
        # the topmost animation - but we need to get called back at a certain
        # point in that animation cycle in order to run the edit queue safely -
        # and animations don't know about the Project!
        #
        # So the monkey-patch above.  Ugly.  When we rewrite the
        # animation cycle, it will be freed from the topmost animation, and
        # this problem will go away.

        self.controls = [create(c, 'control') for c in controls]

        for d in self.drivers:
            d.set_project(self)

        self.animation.set_project(self)

    def start(self):
        with self.LOCK:
            if self.running:
                return

            self.running = True
            self.PROJECTS_RUNNING.add(self)

        self.layout.start()
        for c in self.controls:
            c.set_project(self)
            c.start()
        self.animation.start()

    def stop(self):
        with self.LOCK:
            if not self.running:
                return
            self.running = False
            self.PROJECTS_RUNNING.remove(self)

        log.debug('Project %s stop called on pid %d', self, os.getpid())

        self.animation.stop()
        for c in self.controls:
            c.stop()
        self.layout.stop()

    def cleanup(self):
        self.animation.cleanup()
        for c in self.controls:
            c.cleanup()
        self.layout.cleanup_drivers()

    def join(self, timeout=None):
        self.animation.join(timeout)
        self.stop()

        for c in self.controls:
            c.stop_event.wait()
        self.layout.join(timeout)

    def run(self):
        try:
            self.start()
            self.join()
        finally:
            self.cleanup()
        log.debug('Project %s finishes run()', self)

    def time(self):
        return self.clock.time()

    def sleep(self, delta_time):
        # TODO: why is delta_time occasionally tiny and negative?
        return self.clock.sleep(max(delta_time, 0))

    def flat_out(self, time=0):
        self.clock = FlatOutClock(time)

    @staticmethod
    def stop_all():
        for p in list(Project.PROJECTS_RUNNING):
            p.stop()
        Project.PROJECTS_RUNNING.clear()

    PROJECTS_RUNNING = weakref.WeakSet()


def project(*descs, root_file=None):
    """
    Make a new project, using recursion and alias resolution.

    Use this function in preference to calling Project() directly.
    """
    load.ROOT_FILE = root_file

    desc = defaults.merge(*descs)
    path = desc.get('path', '')

    if root_file:
        project_path = os.path.dirname(root_file)
        if path:
            path += ':' + project_path
        else:
            path = project_path

    with load.extender(path):
        desc = recurse.recurse(desc)

    project = construct.construct(**desc)
    project.desc = desc
    return project


class FlatOutClock:
    def __init__(self, _time=0):
        self._time = _time or time.time()

    def time(self):
        return self._time

    def sleep(self, delta_time):
        assert delta_time >= -MIN_TIME
        self._time += delta_time



"""
List all BiblioPixel Animations
"""

import os, bibliopixel, inspect, BiblioPixelAnimations
from bibliopixel import animation
from bibliopixel.animation import (
    collection, indexed, parallel, receiver, wrapper)
from bibliopixel.util import log, walk
from bibliopixel.project import importer


def run(args):
    animations = {}
    for root, filename in walk.walk(ROOTS):
        basepath = os.path.dirname(root)
        relpath = os.path.relpath(filename, basepath)
        for category, name in _get_classes(relpath):
            animations.setdefault(category, []).append(name)

    count = 0
    for category, entries in sorted(animations.items()):
        log.printer(category + ':')
        for e in sorted(entries):
            if e not in _BLACKLIST:
                log.printer(' ', _canonicalize(e))
                count += 1
        log.printer()

    log.printer('Total animation classes', count)


def add_arguments(parser):
    parser.set_defaults(run=run)


def _animation_category(value):
    if not (inspect.isclass(value) and issubclass(value, animation.Animation)):
        return

    for cls_name, cls in BASE_CLASSES:
        if value is cls or value is animation.Animation:
            return

        if issubclass(value, cls):
            return cls_name
    return 'all'


def _get_classes(filename):
    if not filename.endswith('.py'):
        return

    filename = filename[:-3]
    path = []

    while filename:
        filename, segment = os.path.split(filename)
        path.insert(0, segment)

    if not path or path[-1].startswith('_'):
        return

    python_path = '.'.join(path)

    try:
        module = importer.import_module(python_path)
    except:
        return

    python_path += '.'
    for name, value in vars(module).items():
        category = _animation_category(value)
        if category and not name.startswith('_'):
            yield category, python_path + name


def _canonicalize(name):
    parts = name.split('.')
    p = parts[-1].lower().replace('_', '')
    q = parts[-2].lower().replace('_', '')

    if p == q:
        parts.pop()

    name = '.'.join(parts)
    name = name.replace('bibliopixel.animation.', '.')
    name = name.replace('BiblioPixelAnimations.', '$bpa.')
    return name


BP_ROOT = os.path.dirname(bibliopixel.__file__)
BPA_ROOT = os.path.dirname(BiblioPixelAnimations.__file__)
ROOTS = BP_ROOT, BPA_ROOT
BASE_CLASSES = (
    ('circle', animation.Circle),
    ('cube', animation.BaseCubeAnim),
    ('game', animation.BaseGameAnim),
    ('matrix', animation.BaseMatrixAnim),
    ('strip', animation.BaseStripAnim),
    ('parallel', parallel.Parallel),
    ('wrapper', wrapper.Wrapper),
    ('indexed', indexed.Indexed),
    ('collection', collection.Collection),
    ('receiver', receiver.BaseReceiver),
    ('simple', animation.Animation),
)

_BLACKLIST = set((
    'bibliopixel.animation.off.OffAnim',
    'bibliopixel.animation.remote.control.Off',
    'bibliopixel.main.demo_table.Sequence',
    'bibliopixel.main.all_pixel_test.StripChannelTest',
    'bibliopixel.animation.failed.Failed',
    'BiblioPixelAnimations.simple.Fill.Fill',
))



import abc


class Receiver(abc.ABC):
    """
    Several classes receive data and route it to a target, and this is their
    base class, mainly for documentation.

    Receivers must have two methods, `set_project` and `receive`.  `set_project`
    must be called exactly once for each Receiver, and this must be before
    `receive` is ever.
    """
    @abc.abstractmethod
    def set_project(self, project):
        pass

    @abc.abstractmethod
    def receive(self, values):
        pass



import operator


def advance_permutation(a, increasing=True, forward=True):
    """
    Advance a list of unique, ordered elements in-place, lexicographically
    increasing or backward, by rightmost or leftmost digit.

    Returns False if the permutation wrapped around - i.e. went from
    lexicographically greatest to least, and True in all other cases.

    If the length of the list is N, then this function will repeat values after
    N! steps, and will return False exactly once.

    See also https://stackoverflow.com/a/34325140/43839
    """

    if not forward:
        a.reverse()

    cmp = operator.lt if increasing else operator.gt
    try:
        i = next(i for i in reversed(range(len(a) - 1)) if cmp(a[i], a[i + 1]))
        j = next(j for j in reversed(range(i + 1, len(a))) if cmp(a[i], a[j]))
    except StopIteration:
        # This is the lexicographically last permutation.
        if forward:
            a.reverse()
        return False

    a[i], a[j] = a[j], a[i]
    a[i + 1:] = reversed(a[i + 1:])
    if not forward:
        a.reverse()

    return True



import contextlib, queue

"""
Typical usage would look like this.

In the "write thread":

    while write_thread_is_running():
        with double_queue.produce() as input:
           # mutate input queue

In a separate "read thread"

    while read_thread_is_running():
        with double_queue.consume() as output:
           # read output queue

"""


class Queues(object):
    def __init__(self, item, *items):
        self.empty = queue.Queue()
        self.full = queue.Queue()
        self.empty.put(item)
        for i in items:
            self.empty.put(i)

    @contextlib.contextmanager
    def produce(self):
        i = self.empty.get()
        yield i
        self.full.put(i)

    @contextlib.contextmanager
    def consume(self):
        i = self.full.get()
        yield i
        self.empty.put(i)



"""
Find serial devices and update serial device IDs
"""

from .. util import log

CONNECT_MESSAGE = """
Connect just one Serial device (AllPixel) and press enter..."""


def run(args):
    from ..drivers.serial.driver import Serial
    from ..drivers.serial.devices import Devices
    import serial

    run = True
    log.printer("Press Ctrl+C any time to exit.")
    try:
        while run:
            try:
                input(CONNECT_MESSAGE)
                devices = Devices(args.hardware_id, args.baud)
                ports = devices.find_serial_devices()
                if not ports:
                    log.printer("No serial devices found. Please connect one.")
                    continue

                port = sorted(ports.items())[0][1][0]
                id = devices.get_device_id(port)
                log.printer("Device ID of {}: {}".format(port, id))
                newID = input("Input new ID (enter to skip): ")
                if newID != '':
                    try:
                        newID = int(newID)
                        if newID < 0 or newID > 255:
                            raise ValueError()

                        devices.set_device_id(port, newID)
                        id = devices.get_device_id(port)
                        log.printer("Device ID set to: %s" % id)
                    except ValueError:
                        log.printer("Please enter a number between 0 and 255.")
            except serial.SerialException as e:
                log.printer("Problem connecting to serial device. %s" % e)

            except Exception as e:
                log.printer('Programmer error with exception %s' % e)

    except KeyboardInterrupt:
        pass


def add_arguments(parser):
    parser.set_defaults(run=run)
    parser.add_argument('--hardware-id', default='1D50:60AB',
                        help='USB Vendor ID : Product ID of device. '
                             'Defaults to VID:PID for AllPixel')
    parser.add_argument('--baud', default=921600, type=int,
                        help='Serial baud rate.')



"""
An address identifies how to get or set a piece of data within a Python object,
called the "root", using attributes and indexing.

An address description is a string looking like:

::

    .foo.bar[32][5][baz].bang()

which would mean

"given an object "root", the value ``root.foo.bar[32][5]['baz'].bang()``".

Addresses are divided into "segments".

A segment contained in brackets ``[]`` is an index (for a list) or a key (for
a dictionary) - otherwise, it's an attribute.

In the example above, the segments are ``foo``, ``bar``, ``[32]``, ``[5]``,
``[baz]`` and ``bang``; ``foo`` and ``bar`` are attributes.
``baz`` is a string index, and ``32`` and ``5`` are numeric indexes.

You can use an Address to either get or set values in the root object.

Any key that's entirely numeric is taken to be an integer index.  This is
convenient but prevents the creation of dictionaries like ``{1: 'x', '1': 'y'}``
which you probably didn't want to do anyway.
"""

from .. util import data_file, log


def number(s):
    try:
        return data_file.loads(s)
    except:
        return s


class Address:
    class Segment:
        def __init__(self, name):
            self.name = name

        def set(self, root, *value):
            self._set(root, (value[0] if len(value) == 1 else value))

    class Attribute(Segment):
        def get(self, root):
            return getattr(root, self.name)

        def _set(self, root, value):
            setattr(root, self.name, value)

        def __str__(self):
            return '.%s' % self.name

    class Index(Segment):
        def get(self, root):
            return root[self.name]

        def _set(self, root, value):
            root[self.name] = value

        def __str__(self):
            return '[%s]' % self.name

    class Call(Segment):
        def __init__(self):
            pass

        def get(self, root):
            return root()

        def set(self, root, *value):
            root(*value)

        def __str__(self):
            return '()'

    def __init__(self, name=None):
        if not name:
            self.segments = self.assignment = ()
            return

        self.name, *assignment = name.split('=', 1)
        assignment = assignment and assignment[0].strip()

        self.name = self.name.strip()

        try:
            self.segments = list(_generate(self.name))
        except:
            raise ValueError('%s is not a legal address' % name)

        if not self.segments:
            raise ValueError('Empty Addresses are not allowed')

        if not assignment:
            self.assignment = ()
            return

        if isinstance(self.segments[-1], Address.Call):
            raise ValueError('Cannot assign to a call operation')

        self.assignment = tuple(number(s) for s in assignment.split(','))

    def __bool__(self):
        return bool(self.segments)

    def __str__(self):
        return self.name

    @staticmethod
    def _get(root, address):
        for a in address:
            root = a.get(root)
        return root

    def get(self, root):
        return self._get(root, self.segments)

    def set(self, root, *values):
        *first, last = self.segments
        parent = self._get(root, first)
        last.set(parent, *(self.assignment + values))


def _generate(s):
    def extract_calls(p):
        # Extract () pairs from start and finish of a string
        before, after = [], []
        while p.startswith('()'):
            before.append(Address.Call())
            p = p[2:]

        while p.endswith('()'):
            after.append(Address.Call())
            p = p[:-2]

        return before, p, after

    # Split on dots, then use [ and ] to split out indices
    s = s.strip()
    if s.startswith('.'):
        s = s[1:]
    if s.endswith('.'):
        raise ValueError

    for i, part in enumerate(s.split('.')):
        if not part:
            raise ValueError

        head, *rest = part.split('[')

        # If we had e.g. 'xxx()()[yyy]()[zzz]()()'
        # Now we have first='xxx()' and rest = 'yyy]()', 'zzz]()()'
        # They might have written: ()xxx by mistake
        before, head, after = extract_calls(head)

        if before:
            # A call () is only allowed to start the first segment -
            # for example, an address like a.() is forbidden.
            if i or head:
                raise ValueError
            yield from before

        elif head:
            yield Address.Attribute(head)
            yield from after

        elif i:
            # An index [] is only allowed to start the first segment -
            # for example, an address like a.[2] is forbidden.
            raise ValueError

        for r in rest:
            before, r, after = extract_calls(r)
            if before:
                # A segment cannot contain a () - they must all be at top level.
                raise ValueError

            # A segment must have exactly one ']', exactly at the end.
            index, between = r.split(']', 1)
            if between:
                raise ValueError

            index = int(index) if index.isdigit() else index
            yield Address.Index(index)
            yield from after



from . import deprecated

if deprecated.allowed():  # pragma nocover
    class AttributeDict(object):
        """
        A dict that exposes its values as attributes.

        DEPRECATED: use argparse.Namespace, or just a dict.
        """

        def __init__(self, *args, **kwds):
            deprecated.deprecated('util.AttributeDict')
            for k, v in dict(*args, **kwds).items():
                setattr(self, k, v)

        def __setattr__(self, k, v):
            if isinstance(v, dict):
                v = AttributeDict(**v)
            super().__setattr__(k, v)

        def __eq__(self, other):
            return self.__dict__ == other.__dict__

        def __ne__(self, other):
            return self.__dict__ != other.__dict__

        def __bool__(self):
            return bool(self.__dict__)



import math, random
from ... project import importer
from . import segments as _segments

ENVELOPE_PATH = 'bibliopixel.control.envelope.functions'


class Envelope:
    def __init__(self, enabled=True, base_value=0, period=None, phase=0,
                 offset=0, scale=1, loops=0, symmetry=0.5, reverse=False,
                 serpentine=False, power=1, frequency=None, duty_cycle=0.5):
        self.enabled = enabled
        self.base_value = base_value
        self.phase = phase
        self.offset = offset
        self.scale = scale
        self.loops = loops
        self.symmetry = symmetry
        self.duty_cycle = duty_cycle
        self.reverse = reverse
        self.serpentine = serpentine
        self.power = power
        if period:
            self.period = period
        elif frequency:
            self.frequency = frequency
        else:
            self.period = 1

    @property
    def frequency(self):
        return 1 / self.period

    @frequency.setter
    def frequency(self, f):
        self.period = 1 / f

    def _call(self, x):
        raise NotImplementedError

    def __call__(self, delta_time):
        if self.enabled:
            loop = delta_time / self.period
            if not (self.loops and loop >= self.loops):
                x = ((delta_time + self.phase) / self.period) % 1
                if self.reverse:
                    x = 1 - x
                if self.serpentine and loop % 2:
                    x = 1 - x
                if self.power != 1:
                    x = x ** self.power
                x = _apply_skew(x, 1 - self.duty_cycle)
                x = self._call(x)
                x = _apply_skew(x, self.symmetry)
                return x * self.scale + self.offset

        return self.base_value


class Linear(Envelope):
    def _call(self, x):
        return x


class Sine(Envelope):
    def _call(self, x):
        return math.sin(x * 2 * math.pi)


class Triangular(Envelope):
    def _call(self, x):
        return 2 * x if x < 0.5 else 2 - 2 * x


class Square(Envelope):
    def _call(self, x):
        return 0 if x < 0.5 else 1


class Random(Envelope):
    def _call(self, x):
        return random.random()


class Gaussian(Envelope):
    def __init__(self, mean=0.5, stdev=0.25, **kwds):
        super().__init__(**kwds)
        self.stdev = stdev
        self.mean = mean

    def _call(self, x):
        return random.gauss(self.mean, self.stdev)


class Segments(Envelope):
    def __init__(self, segments=(), period=None, **kwds):
        """
        If period is None, then it is set to be the length of the
        segments.

        Each segment is either a single number, indicating a level, or a pair of
        numbers, indicating a level and a time.  Levels without a time are
        assigned the average delay.
        """
        self.segments = _segments.Segments(segments)
        period = period or self.segments.total_time
        super().__init__(period=period, **kwds)

    def _call(self, x):
        return self.segments(x * self.segments.total_time, self.base_value)


def _apply_skew(x, skew):
    if x == skew:
        return 0.5
    elif x < skew:
        return 0.5 * x / skew
    else:
        return 0.5 + 0.5 * (x - skew) / (1 - skew)



import threading, time
from .. util import log


class AnimationThreading:
    """
    AnimationThreading handles threading - and eventually multiprocessing - for
    Animation.
    """

    def __init__(self, runner, run):
        self.runner = runner
        self.run = run
        self.stop_event = threading.Event()
        self.thread = None
        self.frame_overrun = False
        self.sleep = time.sleep

    def set_project(self, project):
        self.project = project
        self.sleep = project.sleep

    def join(self, timeout=None):
        self.thread and self.thread.join(timeout)

    def is_alive(self):
        return self.thread and self.thread.is_alive()

    def cleanup(self, wait=False):
        # if current thread is animation thread this was called
        # by the context manager and thread is therefore already stopped
        if self.thread != threading.current_thread():
            self.stop_event.set()
            if self.thread and wait:
                self.thread.join()

    def target(self):
        is_main = threading.current_thread() is threading.main_thread()
        log.debug('Animation starts on %s thread', 'main' if is_main else 'new')
        self.run()
        log.debug('Animation complete')

    def wait(self, wait_time, timestamps):
        if not wait_time:
            return

        elapsed_time = timestamps[-1] - timestamps[0]
        if elapsed_time >= wait_time + 0.001:
            logger = log.frame if self.frame_overrun else log.warning
            logger('Frame-time of %dms set, but took %dms',
                   1000 * wait_time, 1000 * elapsed_time)
            self.frame_overrun = True

        elif self.runner.threaded:
            self.stop_event.wait(wait_time - elapsed_time)

        else:
            self.sleep(wait_time - elapsed_time)

    def start(self):
        self.stop_event.clear()

        def start_thread(target):
            self.thread = threading.Thread(target=target, daemon=True)
            self.thread.start()

        if self.runner.main:
            if self.runner.threaded:
                start_thread(self.target)
                self.runner.main()

            else:
                start_thread(self.runner.main)
                self.target()

        elif self.runner.threaded:
            start_thread(self.target)

        else:
            self.target()

        self.thread and self.runner.main and self.thread.join()

    def stopped(self):
        # DEPRECATED
        return not self.is_alive()



"""
Cut matrices by row or column and apply operations to them.
"""


class Cutter:
    """
    Base class that pre-calculates cuts and can use them to
    apply a function to the layout.

    Each "cut" is a row or column, depending on the value of by_row.

    The entries are iterated forward or backwards, depending on the
    value of forward.
    """

    def __init__(self, layout, by_row=True):
        self.layout = layout
        cuts = layout.height if by_row else layout.width
        cutter = self.cut_row if by_row else self.cut_column
        self.cuts = [cutter(i) for i in range(cuts)]

    def apply(self, function):
        """
        For each row or column in cuts, read a list of its colors,
        apply the function to that list of colors, then write it back
        to the layout.
        """
        for cut in self.cuts:
            value = self.read(cut)
            function(value)
            self.write(cut, value)


class Slicer(Cutter):
    """
    Implementation of Cutter that uses slices of the underlying colorlist.
    Does not work if the Matrix layout is serpentine or has any reflections
    or rotations.
    """

    def cut_row(self, i):
        return slice(self.layout.width * i, self.layout.width * (i + 1))

    def cut_column(self, i):
        return slice(i, None, self.layout.width)

    def read(self, cut):
        return self.layout.color_list[cut]

    def write(self, cut, value):
        self.layout.color_list[cut] = value


class Indexer(Cutter):
    """
    Slower implementation of Cutter that uses lists of indices and the
    Matrix interface.
    """

    def cut_row(self, i):
        return [(column, i) for column in range(self.layout.width)]

    def cut_column(self, i):
        return [(i, row) for row in range(self.layout.height)]

    def read(self, cut):
        return [self.layout.get(*i) for i in cut]

    def write(self, cut, value):
        for i, v in zip(cut, value):
            self.layout.set(*i, color=v)



def class_name(c):
    """
    :param c: either an object or a class
    :return: the classname as a string
    """
    if not isinstance(c, type):
        c = type(c)

    return '%s.%s' % (c.__module__, c.__name__)



from . animation import Animation
from .. util.colors import make


class Fill(Animation):
    """
    Fill the screen with a single color or a palette of colors
    """
    def __init__(self, *args, color=None, palette=None, **kwds):
        # Handle legacy "color" parameter
        if color:
            if palette:
                raise ValueError('At most one of color and pallete can be set')
            palette = make.colors(color)

        super().__init__(*args, preclear=False, palette=palette, **kwds)

    def step(self, amt=1):
        n = len(self.color_list)
        self.color_list[:] = (self.palette.get(i) for i in range(n))



from ... util import colors
from ... layout import Matrix


def show_image(setter, width, height,
               image_path='', image_obj=None, offset=(0, 0), bgcolor=colors.Off,
               brightness=255):
    """Display an image on a matrix."""
    bgcolor = colors.color_scale(bgcolor, brightness)

    img = image_obj
    if image_path and not img:
        from PIL import Image

        img = Image.open(image_path)
    elif not img:
        raise ValueError('Must provide either image_path or image_obj')

    w = min(width - offset[0], img.size[0])
    h = min(height - offset[1], img.size[1])
    ox = offset[0]
    oy = offset[1]

    for x in range(ox, w + ox):
        for y in range(oy, h + oy):
            r, g, b, a = (0, 0, 0, 255)
            rgba = img.getpixel((x - ox, y - oy))

            if isinstance(rgba, int):
                raise ValueError('Image must be in RGB or RGBA format!')
            if len(rgba) == 3:
                r, g, b = rgba
            elif len(rgba) == 4:
                r, g, b, a = rgba
            else:
                raise ValueError('Image must be in RGB or RGBA format!')

            if a == 0:
                r, g, b = bgcolor
            else:
                r, g, b = colors.color_scale((r, g, b), a)

            if brightness != 255:
                r, g, b = colors.color_scale((r, g, b), brightness)

            setter(x, y, (r, g, b))


def showImage(layout, imagePath="", imageObj=None, offset=(0, 0),
              bgcolor=colors.Off, brightness=255):
    """Display an image on the matrix"""
    if not isinstance(layout, Matrix):
        raise RuntimeError("Must use Matrix with showImage!")

    layout.all_off()

    return show_image(layout.set, layout.width, layout.height, imagePath,
                      imageObj, offset, bgcolor, brightness)


def loadImage(layout, imagePath="", imageObj=None, offset=(0, 0),
              bgcolor=colors.Off, brightness=255):
    """Display an image on the matrix"""

    if not isinstance(layout, Matrix):
        raise RuntimeError("Must use Matrix with loadImage!")

    texture = [[colors.Off for x in range(layout.width)]
               for y in range(layout.height)]

    def setter(x, y, pixel):
        if y >= 0 and x >= 0:
            texture[y][x] = pixel

    show_image(setter, layout.width, layout.height, imagePath, imageObj,
               offset, bgcolor, brightness)

    return texture



import os, sys, serial, time, traceback

from . codes import CMDTYPE, LEDTYPE, SPIChipsets, BufferChipsets
from . devices import Devices
from . import io
from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util import log, util
from ... drivers.return_codes import (
    RETURN_CODES, print_error, raise_error, BiblioSerialError)


class Serial(DriverBase):
    """Main driver for Serial based LED strips and devices like the AllPixel

    Provides the same parameters of
    :py:class:`bibliopixel.drivers.driver_base.DriverBase` as
    well as those below:

    :param ledtype: LED protocol type. One of
        :py:func:`bibliopixel.drivers.ledtype.LEDTYPE`
    :param str dev: Serial device address/path. If left empty, first device
        found will be used.
    :param int spi_speed: SPI datarate for applicable LED types, in MHz
    :param int restart_timeout: Seconds to wait between reconfigure reboot
        and reconnection attempt
    :param int device_id: Device ID to connect to.
    :param str hardwareID: A valid USB VID:PID pair such as "1D50:60AB"
    :param int baudrate: Baud rate to connect to serial device
    """

    def __init__(self, ledtype=None, num=0, dev='',
                 c_order='RGB', spi_speed=2,
                 gamma=None, restart_timeout=3,
                 device_id=None, hardwareID="1D50:60AB",
                 baudrate=921600, **kwds):

        if ledtype is None:
            raise ValueError('Must provide ledtype value!')

        if num == 0:
            raise ValueError('Must provide num value >0!')

        super().__init__(num, c_order=c_order, gamma=gamma, **kwds)
        self.devices = Devices(hardwareID, baudrate)

        if not (1 <= spi_speed <= 24 and ledtype in SPIChipsets):
            spi_speed = 1

        self._spi_speed = spi_speed
        self._com = None
        self._ledtype = ledtype
        self._bufPad = 0
        self.dev = dev
        self.device_version = 0
        self.device_id = device_id
        self._sync_packet = util.generate_header(CMDTYPE.SYNC, 0)

        if self.device_id is not None and not (0 <= self.device_id <= 255):
            raise ValueError("device_id must be between 0 and 255")

        resp = self._connect()
        if resp == RETURN_CODES.REBOOT:  # reboot needed
            log.info(REBOOT_MESSAGE)
            self._close()
            time.sleep(restart_timeout)
            resp = self._connect()
            if resp != RETURN_CODES.SUCCESS:
                raise_error(resp)
            else:
                log.info("Reconfigure success!")
        elif resp != RETURN_CODES.SUCCESS:
            raise_error(resp)

        if type in SPIChipsets:
            log.info("Using SPI Speed: %sMHz", self._spi_speed)

    def cleanup(self):
        """
        SHOULD BE PRIVATE
        """
        if self._com:
            log.info("Closing connection to: %s", self.dev)
            self._close()

    def _connect(self):
        try:
            if not self.dev:
                self.devices.find_serial_devices()
                idv = self.devices.get_device(self.device_id)
                self.device_id, self.dev, self.device_version = idv
            try:
                self._com = serial.Serial(
                    self.dev, baudrate=self.devices.baudrate, timeout=5)
            except serial.SerialException:
                ports = self.devices.devices.values()
                error = "Invalid port specified. No COM ports available."
                if ports:
                    error = ("Invalid port specified. Try using one of: \n" +
                             "\n".join(ports))
                log.info(error)
                raise BiblioSerialError(error)

            packet = util.generate_header(CMDTYPE.SETUP_DATA, 4)
            packet.append(self._ledtype)  # set strip type
            byteCount = self.bufByteCount()
            if self._ledtype in BufferChipsets:
                if self._ledtype == LEDTYPE.APA102 and self.device_version >= 2:
                    pass
                else:
                    self._bufPad = BufferChipsets[
                        self._ledtype](self.numLEDs) * 3
                    byteCount += self._bufPad

            packet.append(byteCount & 0xFF)  # set 1st byte of byteCount
            packet.append(byteCount >> 8)  # set 2nd byte of byteCount
            packet.append(self._spi_speed)
            self._write(packet)

            code = self._read()
            if code is None:
                self.devices.error()

            return code

        except serial.SerialException as e:
            error = ("Unable to connect to the device. Please check that "
                     "it is connected and the correct port is selected.")
            log.error(traceback.format_exc())
            log.error(error)
            raise e

    def set_device_brightness(self, brightness):
        packet = util.generate_header(CMDTYPE.BRIGHTNESS, 1)
        packet.append(self._brightness)
        self._write(packet)
        code = self._read()
        if code == RETURN_CODES.SUCCESS:
            return True
        print_error(code)

    def _send_packet(self):
        if not self._com:
            return

        self._write(self._packet)

        code = self._read()
        if code is None:
            self.devices.error(fail=False)
        elif code != RETURN_CODES.SUCCESS:
            print_error(code)
        else:
            self._flushInput()
            return True

    def _compute_packet(self):
        count = self.bufByteCount() + self._bufPad
        self._packet = util.generate_header(CMDTYPE.PIXEL_DATA, count)

        self._render()
        self._packet.extend(self._buf)
        self._packet.extend([0] * self._bufPad)

    def _send_sync(self):
        self._write(self._sync_packet)

    def _read(self):
        return io.read_byte(self._com)

    def _close(self):
        try:
            return self._com and self._com.close()
        except Exception as e:
            log.error('Serial exception %s in close', e)
        finally:
            self._com = None

    def _write(self, packet):
        try:
            return self._com and self._com.write(packet)
        except Exception as e:
            log.error('Serial exception %s in write', e)

    def _flushInput(self):
        try:
            return self._com and self._com.flushInput()
        except Exception as e:
            log.error('Serial exception %s in flushInput', e)


class TeensySmartMatrix(Serial):
    """Variant of :py:class:`Serial` for use with the Teensy and
    SmartMatrix library. The following provides compatible firmware:
    https://github.com/ManiacalLabs/BiblioPixelSmartMatrix

    All parameters are the same as with :py:class:`Serial`, except the
    default hardwareID is changed to match the Teensy.

    The main difference is that SmartMatrix requires a sync command to keep
    multiple instances of this driver running smoothly.
    """

    def __init__(self, width, height, dev="", device_id=None,
                 hardwareID="16C0:0483", **kwds):
        super().__init__(ledtype=LEDTYPE.GENERIC, num=width * height,
                         device_id=device_id, hardwareID=hardwareID, **kwds)
        self.sync = self._send_sync


REBOOT_MESSAGE = """Reconfigure and reboot needed!
Waiting for controller to restart..."""

from ... util import deprecated
if deprecated.allowed():  # pragma: no cover
    DriverSerial = Serial
    DriverTeensySmartMatrix = TeensySmartMatrix



from . driver_base import DriverBase
from .. util.server_cache import StaticCache
from .. project import construct


class ServerDriver(DriverBase, StaticCache):
    def __init__(self, *args, address, pixel_positions=None, **kwds):
        """
        Args:
            port:  the port on which the server is running.
            pixel_positions:  the positions of the LEDs in 3-d space.
            **kwds:  keywords passed to DriverBase.
        """
        super().__init__(*args, **kwds)
        self.address = address
        self.server = self.thread = None
        self.pixel_positions = pixel_positions
        if pixel_positions:
            self.set_pixel_positions(pixel_positions)

    def start(self):
        self.server = self.cache().get_server(self.address)
        self._on_positions()

    def set_pixel_positions(self, pixel_positions):
        # This is the "automatic" update from the layout - don't overwrite
        # any pixel_position set in the constructor.
        if self.pixel_positions is None:
            self.pixel_positions = pixel_positions
            self._on_positions()

    def _send_packet(self):
        pass

    def _on_positions(self):
        pass

    def cleanup(self):
        self.server and self.server.close()

    def _compute_packet(self):
        self._render()



from . channel_order import ChannelOrder
from .. util.colors import gamma as _gamma
from .. project import attributes, data_maker, fields
import threading, time


class DriverBase(object):
    """
    Base driver class to build other drivers from.

    :param int num: Number of total pixels held by this driver

    :param int width: Width of matrix, for use
        with :py:class:`bibliopixel.layout.matrix.Matrix`

    :param int height: Height of matrix, for use
        with :py:class:`bibliopixel.layout.matrix.Matrix`

    :param str c_order: Color channel order

    :param gamma: Gamma correction table. Preset tables available
        in :py:mod:`bibliopixel.util.colors.gamma`
    """

    # If set_device_brightness is not empty, it's a method that allows you
    # to directly set the brightness for the device.
    #
    # If it is empty, then the brightness is used as a scale factor in rendering
    # the pixels.
    set_device_brightness = None

    pre_recursion = fields.default_converter

    @classmethod
    def construct(cls, project, **desc):
        """Construct a driver from a project and a description."""
        return cls(maker=project.maker, **desc)

    def __init__(self, num=0, width=0, height=0, c_order="RGB",
                 gamma=None, maker=data_maker.MAKER, **kwds):
        attributes.set_reserved(self, 'driver', **kwds)

        if num == 0:
            num = width * height
            if num == 0:
                raise ValueError("Either num, or width and height are needed!")

        self.numLEDs = num
        gamma = gamma or _gamma.DEFAULT
        self.gamma = gamma

        if isinstance(c_order, str):
            c_order = ChannelOrder.make(c_order)

        self.c_order = c_order
        self.perm = ChannelOrder.ORDERS.index(c_order)

        self.pixel_positions = None

        self.width = width
        self.height = height
        self.maker = maker
        self._buf = self._make_buffer()

        self.lastUpdate = 0
        self.brightness_lock = threading.Lock()
        self._brightness = 255
        self._waiting_brightness = None
        self.time = time.time

    def set_pixel_positions(self, pixel_positions):
        """
        Internal Use Only

        Placeholder callback for sending physical pixel layout data to the
        ``SimPixel`` driver.
        """
        pass

    def set_colors(self, colors, pos):
        """
        Use with caution!

        Directly set the pixel buffers.

        :param colors: A list of color tuples
        :param int pos: Position in color list to begin set operation.
        """
        self._colors = colors
        self._pos = pos

        end = self._pos + self.numLEDs
        if end > len(self._colors):
            raise ValueError('Needed %d colors but found %d' % (
                end, len(self._colors)))

    def set_project(self, project):
        self.project = project
        self.time = project.time

    def start(self):
        """
        Called right before this driver will run.  This is the place
        to do things like start threads, not in the constructor.
        """

    def stop(self):
        """
        Called to request any threads or resources to shut down.
        """

    def cleanup(self):
        """
        Called to shut this driver down, and stop all threads and processes.
        """

    def join(self, timeout=None):
        """
        Called to join threads.
        """

    def bufByteCount(self):
        """
        Total number of bytes that the pixel buffer represents.
        Mainly used for drivers such as :py:mod:`bibliopixel.drivers.serial`
        and :py:mod:`.network`
        """
        return 3 * self.numLEDs

    def sync(self):
        """

        The sync() method is called after the entire frame has been
        sent to the device to indicate that it may now be displayed.

        This is particularly useful when there are multiple drivers comprising
        one display which all need to display the next frame at exactly the same
        time.
        """

    def _compute_packet(self):
        """Compute the packet from the colors and position.

        Eventually, this will run on the compute thread.
        """

    def _send_packet(self):
        """Send the packet to the driver.

        Eventually, this will run on an I/O thread.
        """

    def update_colors(self):
        """Apply any corrections to the current color list
        and send the results to the driver output. This function primarily
        provided as a wrapper for each driver's implementation of
        :py:func:`_compute_packet` and :py:func:`_send_packet`.
        """
        start = self.time()

        with self.brightness_lock:
            # Swap in a new brightness.
            brightness, self._waiting_brightness = (
                self._waiting_brightness, None)

        if brightness is not None:
            self._brightness = brightness
            if self.set_device_brightness:
                self.set_device_brightness(brightness)

        self._compute_packet()
        self._send_packet()

        self.lastUpdate = self.time() - start

    def set_brightness(self, brightness):
        """Set the global brightness for this driver's output.

        :param int brightness: 0-255 value representing the desired
            brightness level
        """
        with self.brightness_lock:
            self._waiting_brightness = brightness

    def _render(self):
        """Typically called from :py:func:`_compute_packet` this applies
        brightness and gamma correction to the pixels controlled by this
        driver.
        """
        if self.set_device_brightness:
            level = 1.0
        else:
            level = self._brightness / 255.0
        gam, (r, g, b) = self.gamma.get, self.c_order
        for i in range(min(self.numLEDs, len(self._buf) / 3)):
            c = [int(level * x) for x in self._colors[i + self._pos]]
            self._buf[i * 3:(i + 1) * 3] = gam(c[r]), gam(c[g]), gam(c[b])

    def _make_buffer(self):
        return self.maker.bytes(self.bufByteCount())



import copy
from . import aliases, importer, load


def construct(*args, datatype, typename=None, **kwds):
    """
    Construct an object from a type constructor.

    A type constructor is a dictionary which has a field "datatype" which has
    a callable method to construct a class, and a field "typename" which is the
    Python path of the function or class in "datatype".
    """
    return datatype(*args, **kwds)


def to_type(d):
    fn = load.load_if_filename(d)
    if fn:
        return fn
    if isinstance(d, str):
        return {'typename': d}

    result = copy.deepcopy(d)

    # 'datatype' always overrides 'typename'
    if 'datatype' in result:
        result.pop('typename', None)
    return result


def to_type_constructor(value, python_path=None):
    """"
    Tries to convert a value to a type constructor.

    If value is a string, then it used as the "typename" field.

    If the "typename" field exists, the symbol for that name is imported and
    added to the type constructor as a field "datatype".

    Throws:
         ImportError -- if "typename" is set but cannot be imported
         ValueError -- if "typename" is malformed
    """
    if not value:
        return value

    if callable(value):
        return {'datatype': value}

    value = to_type(value)
    typename = value.get('typename')
    if typename:
        r = aliases.resolve(typename)
        try:
            value['datatype'] = importer.import_symbol(
                r, python_path=python_path)
            del value['typename']
        except Exception as e:
            value['_exception'] = e

    return value


def construct_type(desc, python_path=None):
    tc = to_type_constructor(desc, python_path)
    return construct(**tc)



from distutils.version import LooseVersion
from .. util import AttributeDict, generate_header, log
import serial, serial.tools.list_ports
from .. drivers.return_codes import RETURN_CODES, print_error
from . gamepad import BaseGamePad

VERSION_ERROR = """
pyserial v{} found, please upgrade to v2.7+!
    pip install pyserial --upgrade
"""

UNABLE_TO_CONNECT_ERROR = """
Unable to connect to the device.
Please check that it is connected and the correct port is selected.
"""

if LooseVersion(serial.VERSION) < LooseVersion('2.7'):
    error = VERSION_ERROR.format(serial.VERSION)
    log.error(error)
    raise ImportError(error)

BTN_MAP = ('A', 'B', 'SELECT', 'START', 'UP', 'DOWN', 'LEFT', 'RIGHT', 'X', 'Y')


class CMDTYPE:
    INIT = 0
    GET_BTNS = 1
    SET_LEDS = 2
    SET_MODE = 4


class SerialPadError(Exception):
    pass


class SerialGamePad(BaseGamePad):
    devices = []

    def __init__(self, btn_map=BTN_MAP, dev="", hardwareID="1B4F:9206"):
        super().__init__()
        self._map = btn_map
        self._hardwareID = hardwareID
        self._com = None
        self.dev = dev

        resp = self._connect()
        if resp != RETURN_CODES.SUCCESS:
            raise SerialPadError(print_error(resp))

    def cleanup(self):
        if self._com is not None:
            log.info("Closing connection to: %s", self.dev)
            self._com.close()

    def close(self):
        from .. util import deprecated
        deprecated.deprecated('SerialGamePad.close')
        self.cleanup()

    @staticmethod
    def find_serial_devices(hardwareID="1B4F:9206"):
        hardwareID = "(?i)" + hardwareID  # forces case insensitive
        if len(SerialGamePad.devices) == 0:
            SerialGamePad.devices = []
            for port in serial.tools.list_ports.grep(hardwareID):
                SerialGamePad.devices.append(port[0])

        return SerialGamePad.devices

    @staticmethod
    def _printError(error):
        msg = "Unknown error occured."
        if error == RETURN_CODES.ERROR_SIZE:
            msg = "Data packet size incorrect."
        elif error == RETURN_CODES.ERROR_UNSUPPORTED:
            msg = "Unsupported configuration attempted."
        elif error == RETURN_CODES.ERROR_BAD_CMD:
            msg = "Unsupported protocol command. Check your device version."

        log.error("%s: %s", error, msg)
        raise SerialPadError(msg)

    @staticmethod
    def _comError(fail=False):
        error = "There was an unknown error communicating with the device."
        log.error(error)
        if fail:
            raise IOError(error)

    def _connect(self):
        try:
            if(self.dev == ""):
                SerialGamePad.find_serial_devices(self._hardwareID)

                if len(SerialGamePad.devices) > 0:
                    self.dev = SerialGamePad.devices[0]
                    log.info("Using COM Port: %s", self.dev)

            try:
                self._com = serial.Serial(self.dev, timeout=5)
            except serial.SerialException:
                ports = SerialGamePad.find_serial_devices(self._hardwareID)
                error = "Invalid port specified. No COM ports available."
                if len(ports) > 0:
                    error = "Invalid port specified. Try using one of: \n" + \
                        "\n".join(ports)
                log.info(error)
                raise SerialPadError(error)

            packet = generate_header(CMDTYPE.INIT, 0)
            self._com.write(packet)

            resp = self._com.read(1)
            return resp and ord(resp)

        except serial.SerialException as e:
            error = UNABLE_TO_CONNECT_ERROR
            log.exception(e)
            log.error(error)
            raise e

    def getKeys(self):
        bits = 0
        try:
            packet = generate_header(CMDTYPE.GET_BTNS, 0)
            self._com.write(packet)
            resp = self._com.read(1)

            if len(resp) == 0:
                SerialGamePad._comError()
            elif ord(resp) != RETURN_CODES.SUCCESS:
                print_error(ord(resp))
            else:
                resp = self._com.read(2)
                if len(resp) != 2:
                    SerialGamePad._comError()
                else:
                    bits = ord(resp[0]) + (ord(resp[1]) << 8)
        except:
            log.error("IO Error Communicating With Game Pad!")

        index = 0
        result = {}
        for m in self._map:
            result[m] = (bits & (1 << index) > 0)
            index += 1
        return AttributeDict(result)

    def setLights(self, data):
        temp = [(0, 0, 0)] * len(data.keys())
        for key in data:
            i = self._map.index(key)
            if i >= 0 and i < len(temp):
                temp[i] = (data[key])
        leds = []
        for l in temp:
            leds.extend(l)

        packet = generate_header(CMDTYPE.SET_LEDS, len(leds))
        packet.extend(leds)
        self._com.write(packet)
        resp = self._com.read(1)
        if len(resp) == 0:
            SerialGamePad._comError()
        elif ord(resp) != RETURN_CODES.SUCCESS:
            print_error(ord(resp))

    def setLightsOff(self, count):
        data = {}
        num = 0
        for b in self._map:
            data[b] = (0, 0, 0)
            num += 1
            if num >= count:
                break
        self.setLights(data)



class Limit:
    def __init__(self, ratio=1, knee=0, gain=1, enable=True):
        """
        :param float ratio: the compression ratio (1 means no compression).
            ratio should usually between 0 and 1.

        :param float knee: the ratio where the compression starts to kick in.
            knee should usually be 0 <= knee <= ratio

        :param float gain: post limiter output gain. gain should usually be >= 0
        """
        # ratio and knee are properties, because we need to recompute _scale
        self.ratio = ratio
        self.knee = knee
        self.gain = gain
        self.enable = enable

    def __bool__(self):
        return self.enable and (self.gain != 1 or self.ratio < 1)

    def limit(self, value):
        if not (self.enable or self):
            return value

        overshoot = value - self.knee
        if overshoot >= 0:
            scale = (self.ratio - self.knee) / (1 - self.knee or 1)
            value = self.knee + overshoot * scale

        value = min(value, self.ratio)
        return value * self.gain

    def limit_colors(self, colors, math):
        if self:
            total = math.sum(colors)
            max_total = 3 * 255 * len(colors)
            average = total / max_total
            limit = self.limit(average) / average
            math.scale(colors, limit)



import os, time
from . import file_writer, renderer
from .. import colors, log

DEFAULT_RENDER = {
    'color': colors.Black,
    'pixel_width': 12,
    'pixel_height': None,
    'ellipse': True,
    'vertical': False,
    'frame': 2,
    'padding': 2,
}


class MovieWriter:
    """Write an animated GIF given frames from an animation."""

    def __init__(self, filename='output.gif', render=None,
                 divide=1, frames=128, time=10, speed=1.0, options=None,
                 gif_dir=None, duration=None, fps=None):
        self.render = dict(DEFAULT_RENDER, **(render or {}))
        self.divide = divide
        self.frames = frames
        self.time = time
        self.speed = speed
        self.duration = duration
        self.fps = fps
        self.options = options or {}
        self.gif_dir = gif_dir

        filename = os.path.expanduser(os.path.abspath(filename))
        file_root, suffix = os.path.splitext(filename)
        self.suffix = suffix

        self.filename = file_root + suffix
        self.basename = os.path.basename(file_root)
        os.makedirs(os.path.dirname(self.filename), exist_ok=True)

        self.finished = False
        self.stop_after_write = True
        self.times = []

    def set_project(self, project):
        self.project = project
        self.render = renderer.renderer(project.layout, **self.render)
        assert self.render
        self.file_writer = file_writer.file_writer(self)
        self.fps = project.animation.runner.fps

    def step(self, cur_step):
        if self.finished or (self.divide >= 1 and cur_step % self.divide):
            return True

        self.frame = cur_step / max(self.divide, 1)
        self.times.append(self.project.time())

        if self.time:
            elapsed = self.times[-1] - self.times[0]
            self.finished = (elapsed >= self.time)
        else:
            self.finished = (self.frame >= self.frames)

        if self.finished:
            self.write()
        else:
            self.file_writer.step()

        return self.finished

    def write(self):
        self.file_writer.write()

    @property
    def length(self):
        return self.times and self.times[-1] - self.times[0] or 0

    @property
    def fps(self):
        if self._fps:
            return self._fps

        if self.duration:
            return 1 / self.duration

        if len(self.times) < 2:
            return 1

        return (len(self.times) - 1) / self.length

    @fps.setter
    def fps(self, fps):
        self._fps = fps

    @property
    def scaled_fps(self):
        return self.fps * self.speed



import contextlib, flask, pathlib, traceback, urllib
from . import flask_server, decorator
from ... util import data_file


class RestServer:
    def __init__(self, port, external_access, open_page,
                 root_folder, index_file):
        root_folder = pathlib.Path(root_folder)
        static_folder = str(root_folder / 'static')

        index_file = pathlib.Path(index_file)
        if not index_file.is_absolute():
            index_file = root_folder / index_file
        self.index_file = str(index_file)
        self.project = None
        self.port = port
        self.flask_server = flask_server.FlaskServer(
            port, external_access, open_page, static_folder=static_folder)

        route = self.flask_server.app.route

        route('/get/<address>', methods=['GET'])(self.get)
        route('/set/<address>/<value>', methods=['GET'])(self.quoted_put)

        route('/single/<address>', methods=['GET'])(self.get)
        route('/single/<address>', methods=['PUT'])(self.put)

        route('/multi', methods=['GET'])(self.multi_get)
        route('/multi/<address>', methods=['GET'])(self.multi_get)

        route('/multi', methods=['PUT'])(self.multi_put)
        route('/multi/<address>', methods=['PUT'])(self.multi_put)

        self.flask_server.start()

    def set_project(self, project):
        self.project = project

    def index(self):
        with open(self.index_file) as fp:
            return fp.read()

    @decorator.single
    def get(self, editor):
        return self._get(editor)

    @decorator.single
    def quoted_put(self, editor, value):
        unquoted = urllib.parse.unquote_plus(value)
        return self._set(editor, unquoted)

    @decorator.single
    def put(self, editor):
        value = flask.request.values['value']
        return self._set(editor, value)

    @decorator.multi
    def multi_get(self, editor, address):
        return self._get(editor)

    @decorator.multi
    def multi_put(self, editor, address):
        value = flask.request.values[address]
        return self._set(editor, value)

    def _get(self, editor):
        return data_file.dumps(editor.get(), use_yaml=False)

    def _set(self, editor, value):
        loaded = data_file.loads(value)
        editor.set(loaded)
        return True

    @contextlib.contextmanager
    def _abort(self, code):
        try:
            yield
        except:
            if self.verbose:
                traceback.print_exc()
            flask.abort(code)



"""
Send a kill signal to a BiblioPixel process running on this
machine to abruptly kill it

DEPRECATED: use

.. code-block:: bash

    $ kill -kill `bpa-pid`

"""

DESCRIPTION = """
Example:

.. code-block:: bash

    $ bp kill

"""

from . import restart


def add_arguments(parser):
    restart.add_signal_arguments(parser, 'SIGKILL')



from . import load
from . import fields
from distutils.version import LooseVersion

MINIMUM_VERSIONS = {'serial': '2.7'}

INSTALL_NAMES = {
    'BiblioPixelAnimations': 'BiblioPixelAnimations',
    'flask': 'flask',
    'noise': 'noise',
    'serial': 'pyserial',
}


def _validate_typename(typename):
    root_module = typename.split('.')[0]
    min_version = MINIMUM_VERSIONS.get(root_module)
    if not min_version:
        return

    version = __import__(root_module).VERSION

    if LooseVersion(version) < LooseVersion(min_version):  # pragma: no cover
        install_name = INSTALL_NAMES.get(root_module, root_module)
        raise ValueError(VERSION_MESSAGE % (
            root_module, version, min_version, install_name))


def _import(typename, python_path=None, loader=load.code):
    try:
        result = loader(typename, python_path=python_path)
        _validate_typename(typename)
        return result

    except ImportError as e:
        root_module = typename.split('.')[0]
        install_name = INSTALL_NAMES.get(root_module)
        if install_name:
            try:
                __import__(root_module)
            except ImportError:
                msg = MISSING_MESSAGE % (root_module, install_name)
                e.msg = msg + e.msg
        raise


def import_symbol(typename, python_path=None):
    return _import(typename, python_path)


def import_module(typename, python_path=None):
    return _import(typename, python_path, loader=load.module)


def make_object(*args, typename=None, python_path=None, datatype=None, **kwds):
    """Make an object from a symbol."""
    datatype = datatype or import_symbol(typename, python_path)
    field_types = getattr(datatype, 'FIELD_TYPES', fields.FIELD_TYPES)
    return datatype(*args, **fields.component(kwds, field_types))


VERSION_MESSAGE = """
You have version %s of module '%s' but you need version %s.

Please upgrade at the command line with:

    $ pip install %s --upgrade
"""

MISSING_MESSAGE = """
You are missing module '%s'.

Please install it at the command line with:

    $ pip install %s
"""

SERIAL_IS_INSTALLED_NOT_PYSERIAL_MESSAGE = """
You have the module `serial` installed, but you need the package `pyserial`
instead.  Sorry, it's not our fault that there are two packages whose names are
so close!

To uninstall that module and install the correct one from the command line:

    $ pip uninstall -y serial
    $ pip install pyserial
"""



from . animation import Animation
from .. layout import Strip


class BaseStripAnim(Animation):
    LAYOUT_CLASS = Strip
    LAYOUT_ARGS = 'num',

    def __init__(self, layout, start=0, end=-1, **kwds):
        super().__init__(layout, **kwds)

        self._start = max(start, 0)
        self._end = end
        if self._end < 0 or self._end >= self.layout.numLEDs:
            self._end = self.layout.numLEDs - 1

        self._size = self._end - self._start + 1



from . driver_base import DriverBase
from .. util.image import movie_writer as _movie_writer


class MovieWriter(DriverBase):
    """
    An animation that writes animated GIFs for each frame in the contained
    animation."""

    def __init__(self, *args, filename='output.gif', render=None,
                 divide=1, frames=0, time=10, speed=1.0, options=None,
                 gif_dir=None, **kwds):
        """
        :param str filename: Base filename to write the animated GIF file

        :param dict render: Parameters to the renderer function -
            see ``bibliopixel.util.image.render.renderer``

        :param int divide: If greater than 1, only rendered one in ``divide``
            frames

        :param int frames: Number of frames to write

        :param float time: Total time to write.  If non-zero, takes precedence
            over `frames`

        :param float speed: the speed of the GIF is scaled up by this factor,
            so if speed=2 then a 2 second animation will become a 1 second GIF.

        :param dict options: Options to
            ``bibliopixel.util.image.gif.write_animation``

        :param str gif_dir: If set, write individual GIF frame files to this
            directory, and do not delete them when done.  For testing purposes.
        """
        super().__init__(*args, **kwds)
        self.cur_step = 1
        self.movie_writer = _movie_writer.MovieWriter(
            filename, render, divide, frames, time, speed, options, gif_dir)

    def set_project(self, project):
        super().set_project(project)
        self.movie_writer.set_project(project)

    def update_colors(self):
        super().update_colors()
        if self.movie_writer.step(self.cur_step):
            self.project.stop()
        else:
            self.cur_step += 1

    def cleanup(self):
        self.movie_writer.write()



import importlib, sys
from .. import commands

DEFAULT_COMMAND = 'run'
ERROR = 'No command entered!'
USAGE = """\
No command entered!
Valid commands are:

    %s

For help on each command, type
    bp <command> -h
""" % ', '.join(commands.COMMANDS)


def main(argv=None):
    argv = sys.argv[1:] if argv is None else list(argv)

    for i, arg in enumerate(argv):
        if arg in commands.COMMANDS:
            command = argv.pop(i)
            break
    else:
        command = DEFAULT_COMMAND
        if all(a.startswith('-') for a in argv):
            from .. util import log
            log.error(ERROR + USAGE)
            return -1

    module = importlib.import_module('bibliopixel.commands.' + command)
    if hasattr(module, 'main'):
        return module.main(argv)

    description = module.__doc__ + getattr(module, 'DESCRIPTION', '')

    from . args import set_args
    args = set_args(description, argv, module)
    args.run(args)


if __name__ == '__main__':
    sys.exit(main())



"""
Configure the AllPixel or similar module
"""

import signal, sys
from .. drivers.serial.codes import LEDTYPE
from .. drivers.return_codes import BiblioSerialError
from .. util import log

types = [
    [LEDTYPE.GENERIC, 'GENERIC'],
    [LEDTYPE.LPD8806, 'LPD8806'],
    [LEDTYPE.WS2801, 'WS2801 '],
    [LEDTYPE.NEOPIXEL, 'WS2812/NEOPIXEL'],
    [LEDTYPE.APA104, 'APA104'],
    [LEDTYPE.WS2811_400, 'WS2811_400'],
    [LEDTYPE.TM1809, 'TM1809/TM1804'],
    [LEDTYPE.TM1803, 'TM1803 '],
    [LEDTYPE.UCS1903, 'UCS1903'],
    [LEDTYPE.SM16716, 'SM16716'],
    [LEDTYPE.APA102, 'APA102/DOTSTAR'],
    [LEDTYPE.LPD1886, 'LPD1886'],
    [LEDTYPE.P9813, 'P9813']
]


def get_input(msg):
    return input('\n' + msg)


def get_int(msg, invalid=-1):
    v = get_input(msg)
    try:
        return int(v)
    except:
        return invalid


def _show_select_list(msg, values):
    log.printer('\n' + msg)
    shift = len(str(len(values)))
    count = 0
    for v in values:
        log.printer('{}: {}'.format(str(count).rjust(shift), v))
        count += 1
    return get_int('Choice: ')


def run(args):
    from ..drivers.serial.driver import Serial
    from ..drivers.serial.devices import Devices
    from ..drivers.serial.codes import SPIChipsets

    try:
        log.printer('Press Ctrl+C anytime to quit.')

        log.printer('\nScanning for devices...')
        devices = Devices(args.hardware_id, args.baud)
        devs = sorted(devices.find_serial_devices().items())
        d = ''
        if not devs:
            get_input(
                'No devices found! Please connect one and press any key...')
            raise ValueError()
        if len(devs) == 1:
            d = devs[0][1]
        else:
            d = _show_select_list('Select device:', devs)
            if d < 0 or d >= len(devs):
                log.printer('\nInvalid choice!')
                raise ValueError()
            d = devs[d][1]

        t = _show_select_list('Choose LED Type', [v[1] for v in types])
        if t < 0 or t >= len(types):
            log.printer('\nInvalid choice!')
            raise ValueError()
        t = types[t]

        num = get_int('Number of LEDs: ', invalid=0)
        if num <= 0:
            log.printer('\nInvalid choice!')
            raise ValueError()

        spi = 2
        use_spi = t[0] in SPIChipsets
        if use_spi:
            spi = get_int('SPI Speed (1-24): ', invalid=0)
            if spi <= 0:
                log.printer('\nInvalid choice!')
                raise ValueError()

        details = 'Device: {}\nLED Type: {}\nNum LEDs: {}'.format(d, t[1], num)
        if use_spi:
            details += '\nSPI Speed: {}'.format(spi)

        log.printer('\n' + details + '\n')

        try:
            Serial(t[0], num, dev=d, SPISpeed=spi, restart_timeout=6)
            log.printer('\nConfigure complete!')
        except BiblioSerialError:
            log.printer('\nError configuring device!')

    except KeyboardInterrupt:
        sys.exit(128 + signal.SIGINT)

    except ValueError:
        pass


def add_arguments(parser):
    parser.set_defaults(run=run)
    parser.add_argument('--hardware-id', default='1D50:60AB',
                        help='USB Vendor ID : Product ID of device. '
                             'Defaults to VID:PID for AllPixel')
    parser.add_argument('--baud', default=921600, type=int,
                        help='Serial baud rate.')



from enum import IntEnum


def rotate_and_flip(coord_map, rotation, flip):
    rotation = (-rotation % 360) // 90
    for i in range(rotation):
        coord_map = list(zip(*coord_map[::-1]))

    if flip:
        coord_map = coord_map[::-1]

    return coord_map


from ... util import deprecated
if deprecated.allowed():  # pragma: no cover
    class Rotation(IntEnum):
        ROTATE_0 = 0  # no rotation
        ROTATE_90 = 90  # rotate 90 degrees
        ROTATE_180 = 180  # rotate 180 degrees
        ROTATE_270 = 270  # rotate 270 degrees



"""
Fanciful names for integer indexes into lists - either a day of week,
a month, a planet, or a chemical element.


"""

import calendar


def to_index(name):
    if isinstance(name, float):
        raise KeyError('Indexes cannot be floating point')
    try:
        return int(name)
    except:
        pass
    try:
        return NAME_TO_INDEX[name.lower()]
    except:
        raise KeyError('Can\'t understand index "%s"' % name)


def to_names(index):
    return INDEX_TO_NAME[index]


def _combine(*name_lists):
    name_to_index = {}
    index_to_name = {}

    def add(i, name):
        nl = name.lower()
        if nl not in name_to_index:
            name_to_index[nl] = i
            index_to_name.setdefault(i, []).append(name)
        elif nl not in DUPES:
            raise ValueError(name + ' duplicated')

    for z in ZEROES:
        add(0, z)

    for name_list in name_lists:
        for i, names in enumerate(name_list):
            if isinstance(names, str):
                names = names,

            for name in names:
                add(i + 1, name)

    return name_to_index, index_to_name


DUPES = 'may', 'mercury'
ZEROES = 'none', 'nothing', 'zero', 'zip'

DAYS = tuple(zip(calendar.day_abbr, calendar.day_name))
MONTHS = tuple(zip(calendar.month_abbr, calendar.month_name))[1:]
COLORS = 'red', 'orange', 'yellow', 'green', 'blue', 'indigo', 'violet'
PLANETS = ('Mercury', 'Venus', 'Earth', 'Mars', 'Jupiter', 'Saturn', 'Uranus',
           'Neptune', 'Pluto')

ELEMENTS = (
    ('H', 'hydrogen'),
    ('He', 'helium'),
    ('Li', 'lithium'),
    ('Be', 'beryllium'),
    ('B', 'boron'),
    ('C', 'carbon'),
    ('N', 'nitrogen'),
    ('O', 'oxygen'),
    ('F', 'fluorine'),
    ('Ne', 'neon'),
    ('Na', 'sodium'),
    ('Mg', 'magnesium'),
    ('Al', 'aluminum'),
    ('Si', 'silicon'),
    ('P', 'phosphorus'),
    ('S', 'sulfur'),
    ('Cl', 'chlorine'),
    ('Ar', 'argon'),
    ('K', 'potassium'),
    ('Ca', 'calcium'),
    ('Sc', 'scandium'),
    ('Ti', 'titanium'),
    ('V', 'vanadium'),
    ('Cr', 'chromium'),
    ('Mn', 'manganese'),
    ('Fe', 'iron'),
    ('Co', 'cobalt'),
    ('Ni', 'nickel'),
    ('Cu', 'copper'),
    ('Zn', 'zinc'),
    ('Ga', 'gallium'),
    ('Ge', 'germanium'),
    ('As', 'arsenic'),
    ('Se', 'selenium'),
    ('Br', 'bromine'),
    ('Kr', 'krypton'),
    ('Rb', 'rubidium'),
    ('Sr', 'strontium'),
    ('Y', 'yttrium'),
    ('Zr', 'zirconium'),
    ('Nb', 'niobium'),
    ('Mo', 'molybdenum'),
    ('Tc', 'technetium'),
    ('Ru', 'ruthenium'),
    ('Rh', 'rhodium'),
    ('Pd', 'palladium'),
    ('Ag', 'silver'),
    ('Cd', 'cadmium'),
    ('In', 'indium'),
    ('Sn', 'tin'),
    ('Sb', 'antimony'),
    ('Te', 'tellurium'),
    ('I', 'iodine'),
    ('Xe', 'xenon'),
    ('Cs', 'cesium'),
    ('Ba', 'barium'),
    ('La', 'lanthanum'),
    ('Ce', 'cerium'),
    ('Pr', 'praseodymium'),
    ('Nd', 'neodymium'),
    ('Pm', 'promethium'),
    ('Sm', 'samarium'),
    ('Eu', 'europium'),
    ('Gd', 'gadolinium'),
    ('Tb', 'terbium'),
    ('Dy', 'dysprosium'),
    ('Ho', 'holmium'),
    ('Er', 'erbium'),
    ('Tm', 'thulium'),
    ('Yb', 'ytterbium'),
    ('Lu', 'lutetium'),
    ('Hf', 'hafnium'),
    ('Ta', 'tantalum'),
    ('W', 'tungsten'),
    ('Re', 'rhenium'),
    ('Os', 'osmium'),
    ('Ir', 'iridium'),
    ('Pt', 'platinum'),
    ('Au', 'gold'),
    ('Hg', 'mercury'),
    ('Tl', 'thallium'),
    ('Pb', 'lead'),
    ('Bi', 'bismuth'),
    ('Po', 'polonium'),
    ('At', 'astatine'),
    ('Rn', 'radon'),
    ('Fr', 'francium'),
    ('Ra', 'radium'),
    ('Ac', 'actinium'),
    ('Th', 'thorium'),
    ('Pa', 'protactinium'),
    ('U', 'uranium'),
    ('Np', 'neptunium'),
    ('Pu', 'plutonium'),
    ('Am', 'americium'),
    ('Cm', 'curium'),
    ('Bk', 'berkelium'),
    ('Cf', 'californium'),
    ('Es', 'einsteinium'),
    ('Fm', 'fermium'),
    ('Md', 'mendelevium'),
    ('No', 'nobelium'),
    ('Lr', 'lawrencium'),
    ('Rf', 'rutherfordium'),
    ('Db', 'dubnium'),
    ('Sg', 'seaborgium'),
    ('Bh', 'bohrium'),
    ('Hs', 'hassium'),
    ('Mt', 'meitnerium'),
    ('Ds', 'darmstadtium'),
    ('Rg', 'roentgenium'),
    ('Cn', 'copernicium'),
    ('Nh', 'nihonium'),
    ('Fl', 'flerovium'),
    ('Mc', 'moscovium'),
    ('Lv', 'livermorium'),
    ('Ts', 'tennessine'),
    ('Og', 'oganesson'),
)


NAME_TO_INDEX, INDEX_TO_NAME = _combine(DAYS, MONTHS, COLORS, PLANETS, ELEMENTS)



from .. channel_order import ChannelOrder
from . base import SPIBase
from ... util.colors import gamma as _gamma
from . import interfaces


MAX_PIXELS = 455


class WS281X(SPIBase):
    """
    SPI driver for WS2812(b) based LED strips on devices like
    Raspberry Pi, OrangePi, BeagleBone,..

    Provides the same parameters as
    :py:class:`bibliopixel.drivers.SPI.SPIBase`
    """

    def __init__(self, num, gamma=_gamma.WS2812, spi_speed=3.2, **kwargs):
        # WS281x need a base clock of ~1MHz - with the encoding, we need 3 times
        # this clock.  After testing 3.0 to 3.2 looks like a good value
        super().__init__(num, gamma=gamma, spi_speed=spi_speed, **kwargs)
        if isinstance(self._interface, interfaces.SpiFileInterface):
            raise ValueError('SPI File interface is unsupported by WS281X')
        if num > MAX_PIXELS:
            raise ValueError('WS2812X SPI driver only supports {} pixels max.'
                             .format(MAX_PIXELS))

    # WS2812 requires gamma correction so we run it through gamma as the
    # channels are ordered
    # As an hack to emulate PWM we use 3 bit on an SPI interface.
    # data -> spi
    # 0b0  -> 0b100
    # 0b1  -> 0b110
    def _compute_packet(self):
        # self._render()
        super()._compute_packet()

        # buffer for result
        # [0] fixes, first led show green when should be off
        buf2 = bytearray([0])
        for byte in self._buf:
            # save each byte in an int var
            tmp = 0
            for i in range(8):
                tmp |= (0b100 | (0b10 if (byte & 1 << i) > 0 else 0)) << (i * 3)
            # save each byte in little endian
            buf2.append(tmp >> 16 & 0xff)
            buf2.append(tmp >> 8 & 0xff)
            buf2.append(tmp & 0xff)

        self._packet = buf2



def _gen_wheel():
    result = []
    for p in range(385):
        if p < 128:
            r = 127 - p % 128
            g = p % 128
            b = 0
        elif p < 256:
            g = 127 - p % 128
            b = p % 128
            r = 0
        else:
            b = 127 - p % 128
            r = p % 128
            g = 0
        result.append((r, g, b))

    return result


_WHEEL = _gen_wheel()


def wheel_color(position):
    """Get color from wheel value (0 - 384).
    Provided for those used to using it from Adafruit libraries
    """
    return _WHEEL[round(position) % len(_WHEEL)]


def wheel_helper(pos, length, cycle_step):
    """Helper for wheel_color that distributes colors over length and
    allows shifting position."""
    return wheel_color((pos * len(_WHEEL) / length) + cycle_step)



import io, os, pathlib
from .. import log
from . import file_writer


class Writer(file_writer.FileWriter):
    def _write(self, filename, frames, fps, loop=0, palette=256):
        """
        Write a series of frames as a single animated GIF.

        :param str filename: the name of the GIF file to write

        :param list frames: a list of filenames, each of which represents a single
            frame of the animation.  Each frame must have exactly the same
            dimensions, and the code has only been tested with .gif files.

        :param float fps:
            The number of frames per second.

        :param int loop:
            The number of iterations. Default 0 (meaning loop indefinitely).

        :param int palette:
            The number of colors to quantize the image to. Is rounded to
            the nearest power of two. Default 256.
        """

        from PIL import Image
        images = []
        for f in frames:
            data = open(f, 'rb').read()
            images.append(Image.open(io.BytesIO(data)))

        # GIF duration is only measured to a hundredth of a second
        duration = round(1 / fps, 2)
        im = images.pop(0)
        im.save(filename,
                save_all=True,
                append_images=images,
                duration=duration,
                loop=loop,
                palette=palette)
        # See also https://github.com/python-pillow/Pillow/issues/3255


from .. import deprecated
if deprecated.allowed():  # pragma: no cover
    from . old_gif import *  # noqa: F403



import functools, threading


def compose_events(events, condition=all):
    """
    Compose a sequence of events into one event.

    Arguments:
        events:    a sequence of objects looking like threading.Event
        condition: a function taking a sequence of bools and returning a bool.
    """
    events = list(events)
    master_event = threading.Event()

    def changed():
        if condition(e.is_set() for e in events):
            master_event.set()
        else:
            master_event.clear()

    def add_changed(f):
        @functools.wraps(f)
        def wrapped():
            f()
            changed()

        return wrapped

    for e in events:
        e.set = add_changed(e.set)
        e.clear = add_changed(e.clear)

    changed()
    return master_event



from . animation import Animation
from .. layout import Cube


class BaseCubeAnim(Animation):
    LAYOUT_CLASS = Cube
    LAYOUT_ARGS = 'x', 'y', 'z'

    def __init__(self, layout, **kwds):
        super().__init__(layout, **kwds)
        self.x, self.y, self.z = layout.x, layout.y, layout.z



import importlib, os, pathlib
from .. project.importer import import_module
from .. commands import COMMANDS
from .. util import log


BP_HEADER = """
APPENDIX: ``bp <command> --help`` for each command
==================================================
"""

BP_TEMPLATE = """\
``bp {command}``
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

{doc}

{description}
"""

SEPARATOR = """
------------------------------------

"""


def print_help():
    template_file = pathlib.Path(__file__).parent / 'commands.rst.tmpl'

    help_text = template_file.open().read().format(
        command_count=len(COMMANDS),
        commands=[', '.join(COMMANDS[0:8]), ', '.join(COMMANDS[8:])])

    log.printer(help_text)
    log.printer(BP_HEADER)
    for command in COMMANDS:
        module = importlib.import_module('bibliopixel.commands.' + command)
        log.printer(BP_TEMPLATE.format(
            command=command,
            doc=module.__doc__.strip(),
            description=getattr(module, 'DESCRIPTION', '')))


if __name__ == '__main__':
    print_help()



from .. import log
from . import names


def printer(colors, use_hex=False):
    if not colors:
        return

    # Try to flatten the list

    try:
        colors[0][0]
    except:
        assert len(colors) % 3 == 0
        colors = zip(*([iter(colors)] * 3))

    for i, color in enumerate(colors):
        log.printer('%2d:' % i, names.color_to_name(color, use_hex))



import os


def walk(roots):
    if isinstance(roots, str):
        roots = [roots]

    for root in roots:
        for dirpath, dirnames, filenames in os.walk(root):
            for f in filenames:
                yield root, os.path.join(dirpath, f)


def walk_suffix(roots, suffixes):
    for root, filename in walk(roots):
        if any(filename.endswith(s) for s in suffixes):
            yield root, filename



from . import gif


class Writer(gif.Writer):
    def __init__(self, writer):
        writer.gif_dir = writer.gif_dir or writer.basename
        super().__init__(writer)

    def write(self):
        pass



try:
    import curses as _curses
except:
    _curses = None

import itertools, os, queue, weakref
from . driver_base import DriverBase


class Curses(DriverBase):
    """
    A driver that uses the "curses" library to get a colored display on a
    regular terminal.

    Note that this driver only works on some platforms.

    The Linux default terminal probably works, but the exact details depend
    on your specific OS and version.

    The MacOS default (Terminal) works at low frame rates but the colors are
    wrong (I delved into it and we can't work round it).

    However, in the MacOS Term2 program, it works at framerates up to at least
    10fps.

    TODO: check how it works on the Rpi (#803)
    """
    DRIVERS = weakref.WeakSet()

    def __init__(self, *args, column_width=3, char='●', **kwds):
        super().__init__(*args, **kwds)

        if not (self.width or self.height):
            self.width = self.numLEDs
            self.height = 1

        if list(Curses.DRIVERS):
            raise ValueError('More than one Curses driver is not allowed')

        Curses.DRIVERS.add(self)

        self.column_width = column_width
        self.offset = column_width // 2
        self.char = chr(char) if isinstance(char, int) else char
        self.queue = queue.Queue()

    def _send_packet(self):
        self.queue.put(True)

    def stop(self):
        super().stop()
        self.queue.put(False)

    @staticmethod
    def main():
        """
        If a project has a Curses driver, the section "main" in the section
        "run" must be "bibliopixel.drivers.curses.Curses.main".

        """
        if not _curses:
            # https://stackoverflow.com/a/1325587/43839
            if os.name == 'nt':
                raise ValueError('curses is not supported under Windows')
            raise ValueError('Your platform does not support curses.')
        try:
            driver = next(iter(Curses.DRIVERS))
        except:
            raise ValueError('No Curses driver in project')

        _curses.wrapper(driver.run_in_curses)

    def run_in_curses(self, stdscr):
        self.set_curses_colors()
        stdscr.clear()
        stdscr.refresh()
        cy, cx = stdscr.getmaxyx()
        cx //= self.column_width
        height, width = self.height, self.width

        if width > cx:
            # log.warning('Width %d too wide for screen %d', width, cx)
            width = cx

        if height > cx:
            # log.warning('Height %d too tall for screen %d', height, cy)
            height = cy

        while self.queue.get():
            for x in range(width):
                for y in range(height):
                    self.write_color(stdscr, x, y)

            stdscr.refresh()

    def write_color(self, stdscr, x, y):
        red, green, blue = self._colors[self._pos + x + self.width * y]
        r, g, b = red >> 5, green >> 5, blue >> 6
        index = r + 8 * (g + 8 * b)
        pair = _curses.color_pair(index)
        screen_x = self.column_width * x + self.offset

        # Pair 0 always has foreground white; just draw a background if so.
        char = self.char if index else ' '
        stdscr.addch(y, screen_x, char, pair)

    @staticmethod
    def set_curses_colors():
        s = 1000 / 255
        for r, g, b in itertools.product(range(8), range(8), range(4)):
            index = r + 8 * (g + 8 * b)
            # Color 0 is black and can't be changed.
            # Pair 0 is white on black and can't be changed.
            if index:
                r, g, b = r << 5, g << 5, b << 6
                _curses.init_color(index, int(r * s), int(g * s), int(b * s))
                _curses.init_pair(index, index, 0)



import copy
from . make import colors_no_palette as _colors
from bibliopixel.colors import conversions

"""
Map names to palettes
==========================

First, the name is looked up in the mutable ``PROJECT_PALETTES`` dictionary.
The project set ``PROJECT_PALETTES`` from its "palettes" Section.

If that fails, the name is looked up in a fixed, immutable dictionary
``BUILT_IN_PALETTES``.
"""


def get(name=None):
    """
    Return a named Palette, or None if no such name exists.

    If ``name`` is omitted, the default value is used.
    """
    if name is None or name == 'default':
        return _DEFAULT_PALETTE

    if isinstance(name, str):
        return PROJECT_PALETTES.get(name) or BUILT_IN_PALETTES.get(name)


# This global variable is changed by the Project to allow color settings
# in Animations and other classes to use palette names.  TODO: fix this.
# types would somehow need to know about the current Project, non-trivial.
PROJECT_PALETTES = {}

BUILT_IN_PALETTES = {
    # "Classic" palettes from old BP
    'rainbow': _colors(conversions.HUE_RAINBOW),
    'raw': _colors(conversions.HUE_RAW),
    'spectrum': _colors(conversions.HUE_SPECTRUM),
    'three_sixty': _colors(conversions.HUE_360),

    # Custom palettes
    'flag': _colors('red, white, blue'),
    'checker': _colors('white, black', serpentine=True),
    'primary': _colors('red, green, blue'),
    'secondary': _colors('yellow, magenta, cyan'),
    'eight': _colors('black, red, yellow, green, cyan, blue, magenta, white'),
    'van_gogh': _colors('gold, teal, spring green 3, ivory black'),
    'trendy': _colors('medium aquamarine, orange red, chocolate 1, gainsboro'),
    'muted': _colors('navajo white 3, ivory 3, white smoke, ivory black'),
    'bold': _colors('black, dodger blue 4, hot pink 4, orange red 3'),
    'clean': _colors('light cyan 2, dark gray, brown 1, snow 2'),
    'warm': _colors('snow 3, dark grey, indian red 4, burly wood 3'),
    'sharp': _colors('sea green 3, ivory black, burly wood 2, gray'),
    'pastel': _colors('azure 3, eggshell, sienna 2, goldenrod'),
    'tints': _colors('plum 4, dark grey, antique white 2, linen'),
    'splash': _colors('steel blue 3, gainsboro, gold, tan 2'),
    'energy': _colors('magenta, yellow, cyan, beige'),
}


def set_default(palette):
    p = get(palette)
    if not p:
        raise ValueError('Don\'t understand default %s' % palette)
    global _DEFAULT_PALETTE
    _DEFAULT_PALETTE = p


set_default('rainbow')



import math, threading, time

from .. util import colors, deprecated, log
from . import matrix_drawing as md
from . import font
from . layout import MultiLayout
from . geometry import make_matrix_coord_map_multi
from . geometry.matrix import (
    make_matrix_coord_map, make_matrix_coord_map_positions)


ROTATION_WARNING = """
Matrix.rotation must be a multiple of 90 degrees but was in fact %s degress.
It was rounded to %s degrees."""


class Matrix(MultiLayout):
    CLONE_ATTRS = MultiLayout.CLONE_ATTRS + (
        'width', 'height', 'rotation', 'vert_flip', 'y_flip', 'serpentine',
        'pixelSize')

    def __init__(self, drivers, width=0, height=0,
                 rotation=0, vert_flip=False, y_flip=False,
                 serpentine=True,
                 threadedUpdate=False, brightness=255,
                 pixelSize=(1, 1), **kwargs):
        """Main class for matricies.

        driver --  instance that inherits from DriverBase
        width --  X axis size of matrix
        height --  Y axis size of matrix
        coord_map --  a 2D matrix defining the X,Y to strip index mapping.
            Not needed in most cases
        rotation -- how to rotate when generating the map.
            Not used if coord_map specified
        vert_flip - flips the generated map along the Y axis.
            This along with rotation can achieve any orientation
        """
        self.gen_multi = make_matrix_coord_map_multi
        super().__init__(drivers, threadedUpdate, brightness, **kwargs)

        rot_mod = rotation % 360
        self.rotation = 90 * round(rot_mod / 90)
        if self.rotation != rot_mod:
            log.warning(ROTATION_WARNING, rotation, self.rotation)

        self.width = width or getattr(self.drivers[0], 'width') or 32
        self.height = height or getattr(self.drivers[0], 'height') or 32
        self.vert_flip = vert_flip
        self.y_flip = y_flip
        self.serpentine = serpentine
        self.pixelSize = pixelSize
        pw, ph = self.pixelSize

        # If both are 0, try to assume it's a square display.
        if self.width == 0 and self.height == 0:
            square = int(math.sqrt(self.numLEDs))
            if (square * square) == self.numLEDs:
                self.width = self.height = square
            else:
                raise TypeError('No width or height passed but '
                                'the number of LEDs is not a perfect square')

        if self.width * self.height > self.numLEDs:
            raise ValueError(
                'width * height cannot exceed total pixel count! %s * %s > %s'
                % (self.width, self.height, self.numLEDs))

        if not self.coord_map:
            if len(self.drivers) == 1:
                # TODO: this should really go into documentation
                log.debug(
                    'Auto generating coordinate map. Use make_matrix_coord_map '
                    'directly if more control needed.')

                # was switched to y_flip, but need to keep vert_flip available
                y_flip = y_flip or vert_flip

                self.coord_map = make_matrix_coord_map(
                    self.width, self.height,
                    serpentine=serpentine,
                    rotation=rotation,
                    y_flip=vert_flip)

            elif self.drivers:
                raise TypeError(
                    'Must provide coord_map if using multiple drivers!')

        self.set_pixel_positions(
            make_matrix_coord_map_positions(self.coord_map))

        # If rotation is 90 or 270 degrees, dimensions need to be swapped so
        # they match the matrix rotation.
        if rotation in (90, 270):
            w = self.width
            h = self.height
            self.width = h
            self.height = w

        self.texture = None
        self.set = self._setColor

        if pw < 0 or pw > self.width or ph < 0 or ph > self.height:
            raise ValueError(
                'pixelSize must be greater than 0 '
                'and not larger than total matrix')

        if self.width % pw != 0 or self.height % ph != 0:
            raise ValueError(
                'pixelSize must evenly divide into matrix dimensions!')

        if pw == 1 and ph == 1:
            self._set = self.__setNormal
        else:
            self._set = self.__setScaled
            self.width = self.width / pw
            self.height = self.height / ph
            self.numLEDs = self.width * self.height

        self.fonts = font.fonts

    @property
    def shape(self):
        """Returns ``width, height``"""
        return self.width, self.height

    def get(self, x, y):
        """
        Return the pixel color at position (x, y), or Colors.black if that
        position is out-of-bounds.
        """
        try:
            pixel = self.coord_map[y][x]
            return self._get_base(pixel)
        except IndexError:
            return colors.Black

    def set(self, x, y, color):
        """Set the pixel color at position x, y."""
        # The actual implementation of this method is computed at construction
        # time and monkey-patched in from one of self._setTexture,
        # self.__setNormal or self.__setScaled
        raise NotImplementedError

    def get_pixel_positions(self):
        return make_matrix_coord_map_positions(self.coord_map)

    def loadFont(self, name, height, width, data):
        self.fonts[name] = {
            'data': data,
            'height': height,
            'width': width
        }

    def setTexture(self, tex=None):
        if tex is None:
            self.texture = tex
            self.set = self._setColor
            return

        if not isinstance(tex, list):
            raise ValueError('Texture must be a list!')

        if len(tex) != self.height:
            raise ValueError(
                'Given texture is must be {} high!'.format(self.height))

        for r in tex:
            if not isinstance(r, list):
                raise ValueError('Texture rows must be lists!')
            if len(r) != self.width:
                raise ValueError(
                    'Texture rows must be {} wide!'.format(self.width))

        self.texture = tex
        self.set = self._setTexture

    def __setNormal(self, x, y, color):
        try:
            pixel = self.coord_map[y][x]
            self._set_base(pixel, color)
        except IndexError:
            pass

    def __setScaled(self, x, y, color):
        sx = x * self.pixelSize[0]
        sy = y * self.pixelSize[1]

        for xs in range(sx, sx + self.pixelSize[0]):
            for ys in range(sy, sy + self.pixelSize[1]):
                self.__setNormal(xs, ys, color)

    # Set single pixel to Color value
    def _setColor(self, x, y, color=None):
        try:
            self._set(x, y, color or (0, 0, 0))
        except IndexError:
            pass

    def _setTexture(self, x, y, color=None):
        if x >= 0 and y >= 0:
            try:
                self._set(x, y, color or self.texture[y][x])
            except IndexError:
                pass

    def setHSV(self, x, y, hsv):
        color = colors.hsv2rgb(hsv)
        self._set(x, y, color)

    def setRGB(self, x, y, r, g, b):
        color = (r, g, b)
        self._set(x, y, color)

    ##########################################################################
    # Drawing Functions
    # Lovingly borrowed from Adafruit
    # https://github.com/adafruit/Adafruit-GFX-Library/blob/master/Adafruit_GFX.cpp
    ##########################################################################

    def drawCircle(self, x0, y0, r, color=None):
        """
        Draw a circle in an RGB color, with center x0, y0 and radius r.
        """
        md.draw_circle(self.set, x0, y0, r, color)

    def fillCircle(self, x0, y0, r, color=None):
        """
        Draw a filled circle in an RGB color, with center x0, y0 and radius r.
        """
        md.fill_circle(self.set, x0, y0, r, color)

    def drawLine(self, x0, y0, x1, y1, color=None, colorFunc=None, aa=False):
        """
        Draw a between x0, y0 and x1, y1 in an RGB color.

        :param colorFunc: a function that takes an integer from x0 to x1 and
            returns a color corresponding to that point
        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.draw_line(self.set, x0, y0, x1, y1, color, colorFunc, aa)

    # Bresenham's algorithm
    def bresenham_line(self, x0, y0, x1, y1, color=None, colorFunc=None):
        """
        Draw line from point x0, y0 to x1, y1 using Bresenham's algorithm.

        Will draw beyond matrix bounds.
        """
        md.bresenham_line(self.set, x0, y0, x1, y1, color, colorFunc)

    # Xiaolin Wu's Line Algorithm
    def wu_line(self, x0, y0, x1, y1, color=None, colorFunc=None):
        """
        Draw a between x0, y0 and x1, y1 in an RGB color.

        :param colorFunc: a function that takes an integer from x0 to x1 and
            returns a color corresponding to that point
        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.wu_line(self.set, x0, y0, x1, y1, color, colorFunc)

    def drawRect(self, x, y, w, h, color=None, aa=False):
        """
        Draw rectangle with top-left corner at x,y, width w and height h

        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.draw_rect(self.set, x, y, w, h, color, aa)

    def fillRect(self, x, y, w, h, color=None, aa=False):
        """
        Draw a solid rectangle with top-left corner at (x, y), width w and
        height h.

        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.fill_rect(self.set, x, y, w, h, color, aa)

    def fillScreen(self, color=None):
        """Fill the matrix with the given RGB color"""
        md.fill_rect(self.set, 0, 0, self.width, self.height, color)

    def drawRoundRect(self, x, y, w, h, r, color=None, aa=False):
        """
        Draw a rounded rectangle with top-left corner at (x, y), width w,
        height h, and corner radius r

        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.draw_round_rect(self.set, x, y, w, h, r, color, aa)

    def fillRoundRect(self, x, y, w, h, r, color=None, aa=False):
        """
        Draw a rounded rectangle with top-left corner at (x, y), width w,
        height h, and corner radius r

        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.fill_round_rect(self.set, x, y, w, h, r, color, aa)

    def drawTriangle(self, x0, y0, x1, y1, x2, y2, color=None, aa=False):
        """
        Draw triangle with vertices (x0, y0), (x1, y1) and (x2, y2)

        :param aa: if True, use Bresenham's algorithm for line drawing;
            Otherwise use Xiaolin Wu's algorithm
        """
        md.draw_triangle(self.set, x0, y0, x1, y1, x2, y2, color, aa)

    def fillTriangle(self, x0, y0, x1, y1, x2, y2, color=None, aa=False):
        """
        Draw filled triangle with points x0,y0 - x1,y1 - x2,y2

        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.fill_triangle(self.set, x0, y0, x1, y1, x2, y2, color, aa)

    if deprecated.allowed():  # pragma: no cover
        fillTrangle = fillTriangle

    def drawChar(self, x, y, c, color, bg,
                 aa=False, font=font.default_font, font_scale=1):
        """
        Draw a single character c at at (x, y) in an RGB color.
        """
        md.draw_char(self.fonts, self.set, self.width, self.height,
                     x, y, c, color, bg, aa, font, font_scale)

    def drawText(self, text, x=0, y=0, color=None,
                 bg=colors.Off, aa=False, font=font.default_font, font_scale=1):
        """
        Draw a line of text starting at (x, y) in an RGB color.

        :param colorFunc: a function that takes an integer from x0 to x1 and
            returns a color corresponding to that point
        :param aa: if True, use Bresenham's algorithm for line drawing;
            otherwise use Xiaolin Wu's algorithm
        """
        md.draw_text(self.fonts, self.set, text, self.width, self.height,
                     x, y, color, bg, aa, font, font_scale)


if deprecated.allowed():  # pragma: no cover
    LEDMatrix = Matrix



import contextlib, signal

SIGNAL_NUMBERS = {k: v for k, v in signal.__dict__.items()
                  if k.startswith('SIG') and '_' not in k}

SIGNAL_NAMES = {v: k for k, v in SIGNAL_NUMBERS.items()}


@contextlib.contextmanager
def context(**handlers):
    signals = {}

    def handler(signum, frame):
        signame = SIGNAL_NAMES[signum]
        signals[signame] = True
        handler = handlers[signame]
        handler()

    for signame in handlers:
        if signame in SIGNAL_NUMBERS:  # Windows doesn't have all signals
            signal.signal(SIGNAL_NUMBERS[signame], handler)

    try:
        yield signals

    finally:
        for signame in handlers:
            if signame in SIGNAL_NUMBERS:  # Windows doesn't have all signals
                signal.signal(SIGNAL_NUMBERS[signame], signal.SIG_DFL)



import importlib, itertools, os, tempfile

MODULE = '.'.join(__name__.split('.')[:-1])
COLORS_PER_LINE = 12
SUFFIXES = '.directory', '.gif', '.mp4', '.yml'


def file_writer(movie_writer):
    suffix = movie_writer.suffix or '.directory'
    if suffix not in SUFFIXES:
        raise ValueError('Cannot write %s files' % suffix)

    module = importlib.import_module(MODULE + suffix)
    return module.Writer(movie_writer)


class FileWriter:
    """Base for classes that write a specific type of movie file"""
    def __init__(self, writer):
        self.writer = writer
        self.frame_files = []
        if self.writer.gif_dir:
            self.gif_dir = self.writer.gif_dir
            os.makedirs(self.gif_dir, exist_ok=True)
        else:
            self.tmp_holder = tempfile.TemporaryDirectory()
            self.gif_dir = self.tmp_holder.name

    def step(self):
        frame_name = '%s%04d.png' % (self.writer.basename, self.writer.frame)
        filename = os.path.join(self.gif_dir, frame_name)

        self.writer.render().save(filename)
        self.frame_files.append(filename)

    def write(self):
        fps = self.writer.scaled_fps
        ff, self.frame_files = self.frame_files, []
        if ff:
            self._write(self.writer.filename, ff, fps, **self.writer.options)



from . driver_base import DriverBase
import time


class Dummy(DriverBase):
    """For Testing: Provides no ouput, just a valid interface"""

    def __init__(self, num=1024, delay=0, **kwds):
        """
        Args
            delay: time to wait in seconds to simulate actual hardware
            interface time
        """
        super().__init__(num)
        self._kwds = kwds
        self._delay = delay

    # Push new data to strand
    def _compute_packet(self):
        if self._delay > 0:
            self.project.sleep(self._delay)
        else:
            pass


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    DriverDummy = Dummy



import collections, functools, sys, threading
from . extractor import Extractor
from .. project import construct, importer
from .. util.log_errors import LogErrors
from .. util import deprecated, flatten, log
from .. util.threads import runnable
from . routing import ActionList, Routing


class Control(runnable.Runnable):
    DEFAULT = {'datatype': ActionList}

    def __init__(self, routing=None, default=None, errors='raise',
                 python_path='bibliopixel.control', verbose=False,
                 pre_routing=None):
        """
        :param Address pre_routing: This Address is set with with the message
            after the message is received and converted, but before it is
            routed.
        :param errors: either a number, indicating how many errors to report
           before ignoring them, or one of these strings:
           'raise', meaning to raise an exception
           'ignore', meaning to ignore all errors
           'report', meaning to report all errors
        """
        super().__init__()
        self.verbose = verbose
        self.receive = LogErrors(self._receive, errors)
        default = dict(self.DEFAULT, **(default or {}))
        self.pre_routing = ActionList(pre_routing)
        self.routing = Routing(routing or {}, default or {}, python_path)

    def set_project(self, project):
        self.project = project
        self.pre_routing.set_project(project)
        self.routing.set_project(project)

    def start(self):
        super().start()
        if self.verbose:
            log.info('Starting %s', self)

        self.thread = self._make_thread()
        getattr(self.thread, 'start', lambda: None)()

    def stop(self):
        super().stop()
        getattr(self.thread, 'stop', lambda: None)()

    def _receive(self, msg):
        """
        Receive a message from the input source and perhaps raise an Exception.
        """
        msg = self._convert(msg)
        if msg is None:
            return

        str_msg = self.verbose and self._msg_to_str(msg)
        if self.verbose and log.is_debug():
            log.debug('Message %s', str_msg)

        if self.pre_routing:
            self.pre_routing.receive(msg)

        receiver, msg = self.routing.receive(msg)
        if receiver:
            receiver.receive(msg)
            if self.verbose:
                log.info('Routed message %s (%s) to %s', str_msg[:128], msg,
                         repr(receiver))

    def _convert(self, msg):
        """
        Convert the message to a Control-specific format
        """
        raise NotImplementedError

    def _make_thread(self):
        """
        Returns a new thread to run the loop for this control source.
        """
        pass

    def _msg_to_str(self, msg):
        if msg is None:
            return '(None)'
        return '.'.join(str(s) for s in msg.values()) or '.'

    def __bool__(self):
        return bool(self.routing or self.pre_routing)

    if deprecated.allowed():
        set_root = set_project

        @property
        def root(self):
            return self.project


class ControlLoop:
    """Mixin class for looping controls"""
    def _receive_all_messages(self):
        for msg in self.messages():
            self.receive(msg)
            if not self.running:
                return

    def messages(self):
        """Should yield a sequence of messages from the input source."""
        raise NotImplementedError

    def _make_thread(self):
        return threading.Thread(target=self._receive_all_messages, daemon=True)


class ExtractedControl(Control):
    EXTRACTOR = {}

    def __init__(self, extractor=None, **kwds):
        super().__init__(**kwds)
        extractor = dict(self.EXTRACTOR, **(extractor or {}))
        self.extractor = Extractor(**extractor)

    def _convert(self, msg):
        return self.extractor.extract(msg)


class ExtractedLoop(ExtractedControl, ControlLoop):
    def _make_thread(self):
        return ControlLoop._make_thread(self)



from . driver_base import DriverBase
from .. util import log

try:
    from phue import Bridge
except:
    error = "Unable to import phue. Please install. pip install phue"
    log.error(error)
    raise ImportError(error)

import colorsys


class Hue(DriverBase):
    """
    Driver for interacting with Philips Hue lights.

    Provides the same parameters of :py:class:`.driver_base.DriverBase` as
    well as those below:

    :param str ip: Network hostname or IP address of the Hue base.
    :param list nameMap: List of names to map to each pixel index
    """

    def __init__(self, num, ip, nameMap=None, **kwds):
        super().__init__(num, **kwds)

        if nameMap and len(nameMap) != self.numLEDs:
            raise ValueError(
                "nameMap must have the same number of entries as the number "
                "of LEDs.")

        self._bridge = Bridge(ip)
        self._bridge.connect()
        self._transitionTime = 0

        if nameMap:
            self._lights = self._bridge.get_light_objects('name')
            self._ids = nameMap
        else:
            self._lights = self._bridge.get_light_objects('id')
            self._ids = [l for l in self._lights]

        if len(self._lights) < self.numLEDs:
            raise ValueError(
                "More LEDs were specified than there are available Hue lights.")

        self.setTransitionTime(0)  # start with no transition time

    def setTransitionTime(self, time):
        if time < 0.0 or time > 30.0:
            raise ValueError(
                "Transition time must be between 0.0 and 30.0 seconds.")

        self._transitionTime = int(time * 10)

    def _mapRange(self, value, minFrom, maxFrom, minTo, maxTo):
        return minTo + (maxTo - minTo) * (value - minFrom) / (maxFrom - minFrom)

    def _rgb2hs(self, rgb):
        r = rgb[0] / 255.0
        g = rgb[1] / 255.0
        b = rgb[2] / 255.0

        h, s, v = colorsys.rgb_to_hsv(r, g, b)

        h = int(self._mapRange(h, 0.0, 1.0, 0, 65535))
        s = int(self._mapRange(s, 0.0, 1.0, 0, 254))
        return (h, s)

    def _send_packet(self):
        for i in range(len(self._ids)):
            h, s = self._rgb2hs(self._colors[i + self._pos])
            bri = min(254, self._brightness)
            if s == 0:
                bri = 0

            cmd = {'on': s != 0, 'bri': bri, 'hue': h, 'saturation': s,
                   'transitiontime': self._transitionTime}
            self._bridge.set_light(self._ids[i], cmd)


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    DriverHue = Hue



import functools
from . classic import Black


class Palette(list):
    """
    Palette is a list of one or more colors used to render Animations.

    The method ``Palette.get()`` takes a position in the pallete and returns
    a color.
    """

    def __init__(self, colors=(), continuous=False, serpentine=False, scale=1,
                 offset=0, autoscale=False, length=None):
        """
        Arguments:
            colors: an iterable of colors

            continuous: if True, interpolate linearly between colors; if False,
              use the nearest color from the original list

            serpentine: if True, palette colors are used in reverse order every
              other iteration, giving a back-and-forth effect.  If False,
              palette colors always restart on each iteration

            scale: Scales the incoming index ``i``.  As ``i`` moves from 0
              to ``len(colors) - 1``, the whole palette repeats itself
              ``self.scale`` times

            offset: offset to the incoming index ``i``, applied after scaling

            autoscale: If True, automatically rescale the Palette size to
              match the length of the output.  ``autoscale`` happens before
              ``scale``, so the two work well together to give banding or
              striping effects across your display

           ``length``:
             The length of the output color_list.  If None, use the length of
             the palette itself.  If autoscale=True, ``length`` is used to scale
             the palette to match the output.
        """
        super().__init__(colors)
        if not self:
            self.append(Black)

        self.continuous = continuous
        self.serpentine = serpentine
        self.scale = scale
        self.offset = offset
        self.autoscale = autoscale
        self.length = length

    def __call__(self, position=0):
        return self.get(position)

    def get(self, position=0):
        """
        Return a color interpolated from the Palette.

        In the case where continuous=False, serpentine=False, scale=1,
        autoscale=False, and offset=0, this is exactly the same as plain old []
        indexing, but with a wrap-around.

        The constructor parameters affect this result as documented in the
        constructor.

        Arguments:
           ``position``:
             May be any integer or floating point number
        """
        n = len(self)
        if n == 1:
            return self[0]

        pos = position

        if self.length and self.autoscale:
            pos *= len(self)
            pos /= self.length

        pos *= self.scale
        pos += self.offset

        if not self.continuous:
            if not self.serpentine:
                return self[int(pos % n)]

            # We want a color sequence of length 2n-2
            # e.g. for n=5: a b c d | e d c b | a b c d ...
            m = (2 * n) - 2
            pos %= m
            if pos < n:
                return self[int(pos)]
            else:
                return self[int(m - pos)]

        if self.serpentine:
            pos %= (2 * n)
            if pos > n:
                pos = (2 * n) - pos
        else:
            pos %= n

        # p is a number in [0, n): scale it to be in [0, n-1)
        pos *= n - 1
        pos /= n

        index = int(pos)
        fade = pos - index
        if not fade:
            return self[index]

        r1, g1, b1 = self[index]
        r2, g2, b2 = self[(index + 1) % len(self)]
        dr, dg, db = r2 - r1, g2 - g1, b2 - b1

        return r1 + fade * dr, g1 + fade * dg, b1 + fade * db

    def __eq__(self, other):
        return (isinstance(other, Palette) and
                super().__eq__(other) and
                vars(self) == vars(other))

    def __ne__(self, other):
        return not (self == other)



from .. util import class_name, log


class LogErrors:
    """
    Wraps a function call to catch and report exceptions.
    """

    def __init__(self, function, errors):
        """
        :param function: the function to wrap
        :param errors: either a number, indicating how many errors to report
           before ignoring them, or one of these strings:
           'raise', meaning to raise an exception
           'ignore', meaning to ignore all errors
           'report', meaning to report all errors
        """
        assert isinstance(errors, int) or errors in (
            'raise', 'ignore', 'report')
        self.function = function
        self.errors = errors
        self.error_count = 0

    def __call__(self, *args, **kwds):
        """
        Calls `self.function` with the given arguments and keywords, and
        returns its value - or if the call throws an exception, returns None.
        """
        try:
            return self.function(*args, **kwds)
        except Exception as e:
            self.error_count += 1
            if self.errors == 'raise':
                raise
            if self.errors == 'ignore':
                return
            args = (class_name.class_name(e),) + e.args

        if self.errors == 'report' or self.error_count <= self.errors:
            log.error(str(args))

        elif self.error_count == self.errors + 1:
            log.error('Exceeded errors of %d', self.errors)



import os, time, traceback
import multiprocessing as mp
from .. import log

DEFAULT_COUNT = 4

"""
Distribute tasks amongst workers
Originally from:  https://github.com/rec/bbcprc/blob/master/bbcprc/worker.py
"""


class Worker(mp.Process):
    def __init__(self, queue, counter):
        super().__init__()
        self.time = time.time()
        self.queue = queue
        self.counter = counter

    def run(self):
        for command, *args in iter(self.queue.get, None):
            try:
                command(*args)
            except Exception as e:
                name = getattr(command, '__name__', str(command))
                log.printer('ERROR on command', name, args, e)
                traceback.print_exc()
            else:
                counter = self._increment_counter()
                self._log(command, args, counter)

    def _increment_counter(self):
        with self.counter.get_lock():
            self.counter.value += 1
            return self.counter.value

    def _log(self, command, args, counter):
        delta_t = time.time() - self.time
        average = delta_t / counter

        command = getattr(command, '__name__', str(command))
        msg = '{command}{args} -> {counter} ({average})'
        log.debug(msg.format(**locals()))


class Workers:
    def __init__(self, count=DEFAULT_COUNT):
        self.queue = mp.Queue()
        self.counter = mp.Value('i')
        self.workers = [Worker(self.queue, self.counter) for i in range(count)]

    def __enter__(self):
        for w in self.workers:
            w.start()
        return self

    def __exit__(self, type, value, traceback):
        for w in self.workers:
            self.queue.put(None)

    def run(self, *args):
        self.queue.put(args)


def work_on(function, items, count=DEFAULT_COUNT):
    with Workers(count) as workers:
        for item in items:
            workers.run(function, *item)



#!/usr/local/bin/python3
# https://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html

import cv2
import argparse
import os

# Construct the argument parser and parse the arguments
ap = argparse.ArgumentParser()
ap.add_argument("-ext", "--extension", required=False, default='png', help="extension name. default is 'png'.")
ap.add_argument("-o", "--output", required=False, default='output.mp4', help="output video file")
args = vars(ap.parse_args())

# Arguments
dir_path = '.'
ext = args['extension']
output = args['output']

images = []
for f in os.listdir(dir_path):
    if f.endswith(ext):
        images.append(f)

# Determine the width and height from the first image
image_path = os.path.join(dir_path, images[0])
frame = cv2.imread(image_path)

cv2.imshow('video', frame)
height, width, channels = frame.shape

# Define the codec and create VideoWriter object
fourcc = cv2.VideoWriter_fourcc(*'mp4v')  # Be sure to use lower case
out = cv2.VideoWriter(output, fourcc, 20.0, (width, height))

for image in images:

    image_path = os.path.join(dir_path, image)
    frame = cv2.imread(image_path)

    out.write(frame)  # Write out frame to video

    cv2.imshow('video', frame)
    if (cv2.waitKey(1) & 0xFF) == ord('q'):  # Hit `q` to exit
        break

# Release everything if job is finished
out.release()
cv2.destroyAllWindows()

print("The output video is {}".format(output))  # noqa: T001, T002



from . types import (
    channel_order, color, colors, direction, duration, gamma, int_name, ledtype,
    spi_interface)


FIELD_TYPES = {
    'bgcolor': color,
    'c_order': channel_order,
    'channel_order': channel_order,
    'color': color,
    'color1': color,
    'color2': color,
    'colors': colors,
    'count': int_name,
    'direction': direction,
    'duration': duration,
    'func': int_name,
    'gamma': gamma,
    'i': int_name,
    'ledtype': ledtype,
    'num': int_name,
    'offColor': color,
    'onColor': color,
    'palette': colors,
    'spi_interface': spi_interface,
    'time': duration,
}


def type_converter(types):
    def converter(value):
        for key, maker in types.items():
            if key not in value:
                continue
            try:
                value[key] = maker.make(value[key])
            except Exception as e:
                e.args = ('For key=%s value=%s' % (
                    key, value[key]),) + e.args
                raise

        return value

    return converter


default_converter = type_converter(FIELD_TYPES)


def field_value(field, value, field_types):
    if not (field_types and field in field_types):
        return value

    try:
        return field_types[field].make(value)
    except Exception as e:
        e.args += ('for field %s and value %s' % (field, value), )
        raise


def component(comp, field_types):
    return {k: field_value(k, v, field_types) for k, v in comp.items()}



from . driver_base import DriverBase
from .. util.colors import names
from .. util import log
import time


class Text(DriverBase):
    """For testing:  prints colors to terminal"""

    def __init__(self, num=1024, columns=8, max_colors=16, **kwds):
        """
        Args
            delay: time to wait in seconds to simulate actual hardware
            interface time
        """
        super().__init__(num)
        self.columns = columns
        self.max_colors = max_colors

    # Push new data to strand
    def _compute_packet(self):
        count = self.numLEDs
        if self.max_colors:
            count = min(count, self.max_colors)

        for i in range(count):
            if not i % self.columns:
                log.printer()
            hex_color = names.color_to_name(self._colors[i + self._pos], True)
            log.printer(hex_color, ' ', end='')

        if self.max_colors and self.numLEDs > self.max_colors:
            log.printer('...')
        else:
            log.printer('')
        log.printer('--')



import functools, numbers
from . names import COLORS
from . import palette


@functools.singledispatch
def color(c):
    raise ValueError('Don\'t understand color name "%s"' % c, _COLOR_USAGE)


@color.register(tuple)
def _(c):
    if len(c) != 3:
        raise ValueError('Length %d is not 3' % len(c), _COLOR_USAGE)

    if not all(0 <= x <= 255 for x in c):
        raise ValueError(
            'All components must be between 0 and 255', _COLOR_USAGE)

    return c


@color.register(list)
def _(c):
    return color(tuple(c))


@color.register(numbers.Number)
def _(c):
    if not (0 <= c <= 255):
        raise ValueError(
            'All components must be between 0 and 255', _COLOR_USAGE)
    return (c, c, c)


@color.register(str)
def _(c):
    try:
        return COLORS[c]
    except:
        raise ValueError('Don\'t understand color name "%s"' % c, _COLOR_USAGE)


@functools.singledispatch
def colors(c):
    """Return a color Palette"""
    raise ValueError("Don't understand type %s" % type(c), _COLORS_USAGE)


@colors.register(palette.Palette)
def _(c):
    return c


@colors.register(tuple)
@colors.register(list)
def _(c):
    return _colors(c)


@colors.register(str)
def _(s):
    return _colors(s)


@colors.register(dict)
def _(d):
    return _colors(**d)


@colors.register(type(None))
def _(d):
    return _colors()


def to_triplets(colors):
    """
    Coerce a list into a list of triplets.

    If `colors` is a list of lists or strings, return it as is.  Otherwise,
    divide it into tuplets of length three, silently discarding any extra
    elements beyond a multiple of three.
    """
    try:
        colors[0][0]
        return colors
    except:
        pass

    # It's a 1-dimensional list
    extra = len(colors) % 3
    if extra:
        colors = colors[:-extra]
    return list(zip(*[iter(colors)] * 3))


def _colors(colors=None, **kwds):
    from . import palettes

    p = palettes.get(colors)
    if not p:
        return colors_no_palette(colors, **kwds)

    if not kwds:
        return p

    kwds = dict(vars(p), **kwds)
    return palette.Palette(p, **kwds)


def colors_no_palette(colors=None, **kwds):
    """Return a Palette but don't take into account Pallet Names."""
    if isinstance(colors, str):
        colors = colors.split(',')

    colors = (color(c) for c in to_triplets(colors or ()))
    return palette.Palette(colors, **kwds)


_COLOR_USAGE = """
A Color can be initialized with:

* A list of three numbers: [0, 0, 0] or [255, 0, 255].
* A single number which represents a brightness/gray level: 0, 255, 127
* A string:  "red", "yellow", "gold" naming a color from ...colors.COLORS.

All numbers must be in the range [0, 256) - 0 <= x < 256"""

_COLORS_USAGE = """
Colors is a list of colors.  Each color can be:

* A list of three numbers: [0, 0, 0] or [255, 0, 255].
* A single number which represents a brightness/gray level: 0, 255, 127
* A string:  "red", "yellow", "gold" naming a color from ...colors.COLORS.

All numbers must be in the range [0, 256) - 0 <= x < 256"""



import errno
from .. import log


class StaticCache:
    SERVER_CLASS = None
    SERVER_KWDS = {}
    CACHE = None

    @classmethod
    def cache(cls):
        if not cls.CACHE:
            cls.CACHE = ServerCache(cls.SERVER_CLASS, **cls.SERVER_KWDS)
        return cls.CACHE


class ServerCache:
    """
    A class that caches servers by key so you don't keep closing and re-opening
    the same server and interrupting your connection.

    The exact nature of the key depends on the sort of server.
    For example, for a server socket like SimPixel, it would be just a port
    number, whereas for a UDP connection like Art-Net, it would be a
    port, ip_address pair.
    """

    def __init__(self, constructor, **kwds):
        """
        :param constructor: a function which takes a key and some keywords,
            and returns a new server
        :param kwds: keywords to the ``constructor`` function
        """
        self.servers = {}
        self.constructor = constructor
        self.kwds = kwds

    def get_server(self, key, **kwds):
        """
        Get a new or existing server for this key.

        :param int key: key for the server to use
        """
        kwds = dict(self.kwds, **kwds)
        server = self.servers.get(key)
        if server:
            # Make sure it's the right server.
            server.check_keywords(self.constructor, kwds)
        else:
            # Make a new server
            server = _CachedServer(self.constructor, key, kwds)
            self.servers[key] = server

        return server

    def close(self, key):
        server = self.servers.pop(key, None)
        if server:
            server.server.close()
            return True


class _CachedServer:
    def __init__(self, constructor, key, kwds):
        try:
            self.server = constructor(key, **kwds)
        except OSError as e:
            if e.errno == errno.EADDRINUSE:
                e.strerror += ADDRESS_IN_USE_ERROR.format(key)
                e.args = (e.errno, e.strerror)
            raise
        self.key = key
        self.constructor = constructor
        self.kwds = kwds

    def check_keywords(self, constructor, kwds):
        if self.constructor != constructor:
            raise ValueError(CACHED_SERVER_ERROR.format(
                key=self.key,
                new_type=str(constructor),
                old_type=str(self.constructor)))

        if self.kwds != kwds:
            log.warning(CACHED_KWDS_WARNING.format(server=self, kwds=kwds))

    def close(self):
        pass

    def __getattr__(self, key):
        # Pass through all other attributes to the server.
        return getattr(self.server, key)


ADDRESS_IN_USE_ERROR = """

Cached server {0} on your machine is already in use.
Perhaps BiblioPixel is already running on your machine?
"""

CACHED_SERVER_ERROR = """
Tried to open server of type {new_type} on {port}, but there was already
a server of type {old_type} running there.
"""

CACHED_KWDS_WARNING = """
Cached server for {server.port} had keywords {server.kwds},
but keywords {kwds} were requested.
"""



from .. import log
from . import file_writer

try:
    import cv2
except:
    log.error("""\
Could not import the OpenCV Python library - install it with:

    $ pip install opencv-python
""")
    cv2 = None

# https://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html


class Writer(file_writer.FileWriter):
    def write(self, filename, frames, fps, show=False):
        fps = max(1, fps)
        out = None

        try:
            for image in frames:
                frame = cv2.imread(image)
                if show:
                    cv2.imshow('video', frame)

                if not out:
                    height, width, channels = frame.shape
                    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
                    out = cv2.VideoWriter(filename, fourcc, 20, (width, height))

                out.write(frame)

        finally:
            out and out.release()
            cv2.destroyAllWindows()



from enum import IntEnum
from .. ledtype import LEDTYPE


class CMDTYPE(IntEnum):
    SETUP_DATA = 1  # config data (LEDTYPE, SPI speed, num LEDs)
    PIXEL_DATA = 2  # raw pixel data will be sent as [R1,G1,B1,R2,G2,B2,...]

    # For BRIGHTNESS, data will be single 0-255 brightness value, and
    # length must be 0x00,0x01
    BRIGHTNESS = 3

    GETID = 4
    SETID = 5
    GETVER = 6
    SYNC = 7


SPIChipsets = [
    LEDTYPE.LPD8806,
    LEDTYPE.WS2801,
    LEDTYPE.SM16716,
    LEDTYPE.APA102,
    LEDTYPE.P9813
]

# Chipsets here require extra pixels padded at the end
# Key must be an LEDTYPE
# value a lambda function to calc the value based on numLEDs
BufferChipsets = {
    LEDTYPE.APA102: lambda num: (int(num / 64.0) + 1)
}
